# This code was first implemented as a standard nelder mead (Adapted from Scholorpedia http://dx.doi.org/10.4249/scholarpedia.2928) which was then checked against the amoeba algorithm in "Numerical Recipes in C" by Press et al. The resulting Nelder Mead algorithm was verified against scipy nelder mead, then adjusted to reproduce the simulated annealing of "Numerical Recipes in C" by perturbing the costs.

import numpy as np
import ionbench


# noinspection PyShadowingNames
def run(bm, x0=None, tempInitial=None, N=5, maxIter=1000, debug=False):
    """
    Simulated Annealing by Vanier 1999 uses the algorithm presented in "Numerical Recipes in C" by Press et al. This features a nelder mead optimiser, in which each point in the simplex, in addition to the cost also stored a realisation of a logarithmically distributed random variable. When points are compared, if the point is already in the simplex, this random variable is added to the cost, and if the point is not yet in the simplex (or was added during this step), then it is subtracted from the cost. This ensures that 1) the point which is moved (the worst point in nelder mead) is now random, and 2) there is a non-zero probability to accept a move that increases the cost. As the temperature decreases, the magnitude of this noise also decreases, ensuring the algorithm approaches a standard nelder mead simplex as the number of iterations approaches the maximum number.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. Population is generated by randomly perturbing this initial guess +-50%, then applying the appropriate bounds. If x0=None (the default), then the population will be sampled uniformly between the bounds.
    tempInitial : float, optional
        Set to 1 in Vanier 1999, but changes with the scale of the cost function/problem. If no value is set, the default, then the cost of x0 is used.
    N : float, optional
        Number of moves per temperature update. Vanier 1999 varies this problem specifically, between 5 and 300. The default is 5.
    maxIter : int, optional
        Maximum number of iterations. The default is 1000.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters in the final simplex.

    """
    cost = ionbench.utils.cache.get_cached_cost(bm)

    # noinspection PyShadowingNames
    def wrap_around_boundaries(xOld):
        x = np.copy(xOld)
        if not bm.in_parameter_bounds(bm.original_parameter_space(x)):
            for i in range(bm.n_parameters()):
                ub = bm.input_parameter_space(bm.ub)
                lb = bm.input_parameter_space(bm.lb)
                step = ub[i] - lb[i]
                while x[i] > ub[i]:
                    x[i] -= step
                while x[i] < lb[i]:
                    x[i] += step
        return x
    # class for points on the simplex

    class Point:
        def __init__(self, x):
            self.noise = None
            self.x = wrap_around_boundaries(x)
            self.cost = cost(tuple(x))
            self.regen_noise()

        def regen_noise(self):
            self.noise = -temp * np.log(np.random.rand())

        def tot_cost(self):
            return self.cost + self.noise

    # class for the simplex
    # noinspection PyShadowingNames
    class Simplex:
        def __init__(self, points):
            self.points = points

        def regen_noise(self):
            # Regenerate the random noise for each point in the simplex
            for p in self.points:
                p.regen_noise()

        def get_best(self):
            # returns the best point in the simplex
            x_best = self.points[0]
            for p in self.points:
                if p.tot_cost() < x_best.tot_cost():
                    x_best = p
            return x_best

        def get_worst(self):
            # returns the worst point in the simplex
            x_worst = self.points[0]
            for p in self.points:
                if p.tot_cost() > x_worst.tot_cost():
                    x_worst = p
            return x_worst

        def get_second_worst(self):
            # returns second-worst point in the simplex
            x_worst = self.get_worst()
            x_secondWorst = self.points[0]
            for p in self.points:
                if p.tot_cost() > x_secondWorst.tot_cost() and p != x_worst:
                    x_secondWorst = p
            return x_secondWorst

        def centroid(self, x_worst):
            # find the centroid of the simplex by averaging the x positions, not including x_worst. Don't generate a point, cost not needed
            cen = [0] * bm.n_parameters()
            for p in self.points:
                if p != x_worst:
                    cen += p.x
            cen /= len(self.points) - 1
            return wrap_around_boundaries(cen)

        def accept(self, x):
            x_worst = self.get_worst()
            for i in range(len(self.points)):
                if self.points[i] == x_worst:
                    self.points[i] = x
                    return

        def step(self):
            """
            Take a modified nelder mead step with acceptance probability related to the current temperature.
            """
            x_worst = self.get_worst()
            x_best = self.get_best()
            x_secondWorst = self.get_second_worst()
            c = self.centroid(x_worst)
            # attempt reflection
            xr = Point(2 * c - x_worst.x)
            xr.noise = -xr.noise
            if x_secondWorst.tot_cost() > xr.tot_cost() >= x_best.tot_cost():
                if debug:
                    print("Accept reflection")
                self.accept(xr)
                return
            if xr.tot_cost() < x_best.tot_cost():
                # Attempt expand
                xe = Point(2 * xr.x - c)
                xe.noise = -xe.noise
                if xe.tot_cost() < xr.tot_cost():
                    if debug:
                        print("Accept expansion")
                    self.accept(xe)
                else:
                    if debug:
                        print("Ignore expansion. Accept reflection")
                    self.accept(xr)
                return
            if xr.tot_cost() >= x_secondWorst.tot_cost():
                # contract
                if debug:
                    print("Attempt contraction")
                if x_worst.tot_cost() > xr.tot_cost() >= x_secondWorst.tot_cost():
                    # outside contract
                    xc = Point((xr.x + c) / 2)
                    xc.noise = -xc.noise
                    if xc.tot_cost() <= xr.tot_cost():
                        if debug:
                            print("Accept inside contraction")
                        self.accept(xc)
                        return
                else:
                    # inside contract
                    xc = Point((x_worst.x + c) / 2)
                    xc.noise = -xc.noise
                    if xc.tot_cost() < x_worst.tot_cost():
                        if debug:
                            print("Accept inside contraction")
                        self.accept(xc)
                        return
            # shrink
            if debug:
                print("Shrink")
            for i in range(len(self.points)):
                if self.points[i] != x_best:
                    self.points[i] = Point((x_best.x + self.points[i].x) / 2)
            return

    if x0 is None:
        # sample initial point
        x0 = bm.sample()

    # Initialise the temperature
    if tempInitial is None:
        tempInitial = cost(tuple(x0))
    temp = tempInitial

    # Same initialization procedure as scipy or fminseach
    points = [Point(x0)]
    perturbVector = 0.05 * x0
    perturbVector[perturbVector == 0] = 0.00025
    for i in range(bm.n_parameters()):
        perturb = np.zeros(bm.n_parameters())
        perturb[i] = perturbVector[i]
        points.append(Point(x0 + perturb))

    bestPos = None
    bestCost = np.inf
    for p in points:
        if p.cost < bestCost:
            bestCost = p.cost
            bestPos = np.copy(p.x)

    # Build the simplex
    simplex = Simplex(points)
    i = None
    for i in range(maxIter):
        # Take a modified nelder mead step
        simplex.step()
        # Regenerate the logarithmically distributed noise used in SA modification
        simplex.regen_noise()
        # Annealing schedule
        if i % N == 0:
            if i < 1000:
                temp = tempInitial * (1 - i / 1000)  # Proportional decrease in temperature
            else:
                temp = 0

        for p in simplex.points:
            if p.cost < bestCost:
                bestCost = p.cost
                bestPos = np.copy(p.x)
        if debug:
            print(f'Best cost so far is {bestCost}')
        if bm.is_converged():
            break

    if i >= maxIter-1:
        bm.set_max_iter_flag()

    # Return the best point in the final simplex
    bm.evaluate()
    return bestPos


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Vanier1999

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Vanier1999.

    """
    mod = ionbench.modification.Vanier1999()
    return mod



import numpy as np
import ionbench
import scipy.optimize


# noinspection PyShadowingNames
def run(bm, x0=None, nGens=4000, popSize=None, F=0.5, CR=0.3, debug=False):
    """
    Run differential evolution, as defined by Zhou et al. 2009. This algorithm is based on scheme DE/rand/1 of Storn 1999.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. Population is generated by randomly perturbing this initial guess +-50%, clamped to bounds if necessary. If x0=None (the default), then the population will be sampled using bm.sample().
    nGens : int, optional
        The number of generations to run the optimisation algorithm for. The default is 4000.
    popSize : int, optional
        The size of the population in each generation. The default is None, in which case 10 times the number of parameters will be used.
    F : float, optional
        The weight applied to the difference vector. No recommended value is reported by Zhou et al. 2009. The range is [0.5,1.0] is recommended by Storn 1996 (the citation within Zhou et al. 2009). We have arbitrarily chosen 0.5 to be the default value.
    CR : float, optional
        The crossover probability. No recommended value is reported by Zhou et al. 2009. The value of 0.3 is recommended by Storn 1996 (the citation within Zhou et al. 2009), so we use this as the default value.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    class Individual:
        def __init__(self):
            if x0 is None:
                self.x = bm.sample()
            else:
                self.x = bm.input_parameter_space(bm.original_parameter_space(x0) * np.random.uniform(low=0.5, high=1.5, size=bm.n_parameters()))
            self.x = bm.clamp_parameters(self.x)
            self.cost = None

        def find_cost(self):
            self.cost = cost_func(tuple(self.x))

    # noinspection PyPep8Naming,PyShadowingNames
    def get_L():
        # Get the number of parameters to perturb in crossover
        L = 1
        while np.random.rand() < CR and L < bm.n_parameters():
            L += 1
        return L

    signed_error = ionbench.utils.cache.get_cached_signed_error(bm)
    grad = ionbench.utils.cache.get_cached_grad(bm, residuals=True)

    def cost_func(x):
        """
        Cost function needs to use the signed_error function so that they use the same cache.
        """
        return bm.rmse(signed_error(x) + bm.DATA, bm.DATA)

    # Ensure popSize is defined
    if popSize is None:
        popSize = 10 * bm.n_parameters()

    # Generate initial population
    pop = [None] * popSize
    for i in range(popSize):
        pop[i] = Individual()
        pop[i].find_cost()
    gen = None
    for gen in range(nGens):
        if debug:
            print('----------------')
            print(f'Generation {gen} of {nGens}')
            costMean = 0
            for p in pop:
                costMean += p.cost / popSize
            print(f'Average cost is {costMean}')
        if gen > 0 and gen % 1000 == 0:
            # Run lm on each point
            if debug:
                print('Running lm on each point')
            for i in range(popSize):
                out = scipy.optimize.least_squares(bm.signed_error, pop[i].x, method='lm', jac=grad, max_nfev=1000)
                pop[i].x = out.x
                pop[i].cost = out.cost

        for i in range(popSize):
            # Find trial point
            if debug:
                print(f'Working on individual {i}')
                print(pop[i].x)
            possible = list(range(popSize))
            possible.remove(i)
            r1, r2, r3 = np.random.choice(possible, size=3, replace=False)
            if debug:
                print(f'r1: {r1}, r2: {r2}, r3: {r3}')
            trial = Individual()
            nu = pop[r1].x + F * (pop[r2].x - pop[r3].x)
            if debug:
                print('Perturbed point (nu)')
                print(nu)
            # Do crossover
            trial.x = np.copy(pop[i].x)
            L = get_L()
            n = np.random.choice(range(bm.n_parameters()))
            for j in range(bm.n_parameters()):
                if j == n:
                    for k in range(L):
                        trial.x[(j + k) % bm.n_parameters()] = nu[(j + k) % bm.n_parameters()]
            if debug:
                print(f'Perturbed point after crossover with L {L} and n {n}')
                print(trial.x)
            # Save best out of trial or pop[i]
            try:
                trial.find_cost()
            except Exception as e:  # pragma: no cover
                print(e)
                print('Error in trial.find_cost()')
                print(trial.x)
                raise e
            if trial.cost < pop[i].cost:
                if debug:
                    print(f'Point accepted since trail cost is {trial.cost} and ith cost is {pop[i].cost}')
                pop[i] = trial

        if bm.is_converged():
            break

    if gen >= nGens-1:
        bm.set_max_iter_flag()

    # Find best point in final pop
    bestInd = Individual()
    bestInd.cost = np.inf
    for i in range(popSize):
        if pop[i].cost <= bestInd.cost:
            bestInd = pop[i]
    bm.evaluate()
    return bestInd.x


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Zhou2009

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Zhou2009.

    """
    mod = ionbench.modification.Zhou2009()
    return mod



import numpy as np
import ionbench
import ionbench.utils.population_optimisers as pop_opt
import copy


# noinspection PyShadowingNames
def run(bm, x0=None, nGens=50, eta_cross=10, eta_mut=20, popSize=50, debug=False):
    """
    Runs the genetic algorithm from Bot et al. 2012.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. Population is generated by randomly perturbing this initial guess +-50%, clamped to bounds if necessary. If x0=None (the default), then the population will be sampled using bm.sample().
    nGens : int, optional
        The number of generations to run the optimisation algorithm for. The default is 50.
    eta_cross : float, optional
        Crossover parameter. The default is 10.
    eta_mut : float, optional
        Mutation parameter. The default is 20.
    popSize : int, optional
        The size of the population in each generation. The default is 50.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """

    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    pop = pop_opt.get_pop(bm, x0, popSize, cost_func)
    gen = None
    for gen in range(nGens):
        elites = pop_opt.get_elites(pop, 1)
        if debug:
            print("------------")
            print(f'Gen {gen}, Best cost: {elites[0].cost}')

        # Tournament selection
        pop = pop_opt.tournament_selection(pop)

        # Crossover SBX
        pop = pop_opt.sbx_crossover(pop, bm, cost_func, eta_cross)

        # Mutation
        pop = pop_opt.polynomial_mutation(pop, bm, cost_func, eta_mut)

        if debug:
            print(f'Finishing gen {gen}')

        # Find costs
        pop = pop_opt.find_pop_costs(pop)

        # Elitism
        pop = pop_opt.set_elites(pop, elites)

        if bm.is_converged():
            break

    if gen >= nGens-1:
        bm.set_max_iter_flag()

    elites = pop_opt.get_elites(pop, 1)
    bm.evaluate()
    return elites[0].x


# noinspection PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Bot2012
    modNum = 2 -> Groenendaal2015

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Bot2012.

    """

    if modNum == 1:
        mod = ionbench.modification.Bot2012()
    elif modNum == 2:
        mod = ionbench.modification.Groenendaal2015()
    else:
        mod = ionbench.modification.Empty(name='GA_Bot2012')
    return mod



from ionbench.utils import autodiff
from ionbench.utils.multistart import multistart
from ionbench.utils import population_optimisers
from ionbench.utils import cache
from ionbench.utils import scipy_setup
from ionbench.utils import particle_optimisers
from ionbench.utils import classes_pints
from ionbench.utils import results



from ionbench.optimisers.external_optimisers import GA_Bot2012
from ionbench.optimisers.external_optimisers import GA_Smirnov2020
from ionbench.optimisers.external_optimisers import patternSearch_Kohjitani2022
from ionbench.optimisers.external_optimisers import ppso_Chen2012
from ionbench.optimisers.external_optimisers import hybridPSOTRR_Loewe2016
from ionbench.optimisers.external_optimisers import SA_Vanier1999
from ionbench.optimisers.external_optimisers import SPSA_Spall1998
from ionbench.optimisers.external_optimisers import curvilinearGD_Dokos2004
from ionbench.optimisers.external_optimisers import GA_Cairns2017
from ionbench.optimisers.external_optimisers import GA_Gurkiewicz2007a
from ionbench.optimisers.external_optimisers import GA_Gurkiewicz2007b
from ionbench.optimisers.external_optimisers import DE_Zhou2009
from ionbench.optimisers.external_optimisers import pso_Cabo2022
from ionbench.optimisers.external_optimisers import pso_Seemann2009
from ionbench.optimisers.external_optimisers import NMPSO_Liu2011
from ionbench.optimisers.external_optimisers import NMPSO_Clausen2020
from ionbench.optimisers.external_optimisers import PSO_Loewe2016
from ionbench.optimisers.external_optimisers import PSOTRR_Loewe2016
from ionbench.optimisers.external_optimisers import hybridPSOTRRTRR_Loewe2016
from ionbench.optimisers.external_optimisers import randomSearch_Vanier1999
from ionbench.optimisers.external_optimisers import stochasticSearch_Vanier1999



import ionbench
from ionbench import modification
import numpy as np
from importlib import import_module
import pytest


class TestModifications:
    bm = ionbench.problems.loewe2016.IKr()

    @pytest.mark.cheap
    def test_bounds(self):
        # Check bounds are correctly applied for all settings
        # No bounds initially
        assert not self.bm.parametersBounded
        # Sampler
        mod = modification.Modification(parameterBounds='on')
        mod.apply(self.bm)
        assert self.bm.parametersBounded
        assert all(np.array(self.bm.lb) > -np.inf)
        assert all(np.array(self.bm.ub) < np.inf)
        # Custom
        lb = np.random.rand(self.bm.n_parameters())
        ub = np.random.rand(self.bm.n_parameters()) + 1
        mod = modification.Modification(parameterBounds='Custom', customBounds=[lb, ub])
        mod.apply(self.bm)
        assert self.bm.parametersBounded
        assert all(self.bm.lb == lb)
        assert all(self.bm.ub == ub)
        # None
        mod = modification.Modification()
        mod.apply(self.bm)
        assert not self.bm.parametersBounded

    @pytest.mark.cheap
    def test_rate_bounds(self):
        # Check rate bounds are correctly applied for all settings
        # No rate bounds initially
        assert not self.bm.ratesBounded
        # On
        mod = modification.Modification(rateBounds='on')
        mod.apply(self.bm)
        assert self.bm.ratesBounded
        # Off
        mod = modification.Modification(rateBounds='off')
        mod.apply(self.bm)
        assert not self.bm.ratesBounded
        # By default, they will be turned off
        mod = modification.Modification()
        mod.apply(self.bm)
        assert not self.bm.ratesBounded

    @pytest.mark.cheap
    def test_log_transform(self):
        # Check log transform is correctly applied for all settings
        # No log transform initially
        assert not any(self.bm.logTransformParams)
        # Standard log transform
        mod = modification.Modification(logTransform='on')
        mod.apply(self.bm)
        assert any(self.bm.logTransformParams)
        assert not all(self.bm.logTransformParams)
        # Custom
        transforms = list(np.random.choice([True, False], self.bm.n_parameters() - 2)) + [True, False]  # Random true and falses in which the last two are True and False
        mod = modification.Modification(logTransform='custom', customLogTransform=transforms)
        mod.apply(self.bm)
        assert any(self.bm.logTransformParams)
        assert all(np.array(self.bm.logTransformParams) == np.array(transforms))
        assert not all(self.bm.logTransformParams)
        # No log transform
        mod = modification.Modification()
        mod.apply(self.bm)
        assert not any(self.bm.logTransformParams)

    @pytest.mark.cheap
    def test_scale_factors(self):
        # Check scale factors are correctly applied
        # No scale factors initially
        assert not self.bm.useScaleFactors
        # Turn on scale factors
        mod = modification.Modification(scaleFactors='on')
        mod.apply(self.bm)
        assert self.bm.useScaleFactors
        # Default is off
        mod = modification.Modification()
        mod.apply(self.bm)
        assert not self.bm.useScaleFactors

    @pytest.mark.cheap
    def test_multiple_settings(self):
        # Check multiple settings applied at once doesn't break anything and order in the .apply() method doesn't matter
        mod = modification.Modification(logTransform='on', parameterBounds='on', scaleFactors='On')
        mod.apply_rate_bounds(mod.dict['rateBounds'], self.bm)
        mod.apply_log_transforms(mod.dict['log transform'], self.bm)
        mod.apply_scale_factors(mod.dict['scale factors'], self.bm)
        mod.apply_parameter_bounds(mod.dict['parameterBounds'], self.bm)
        assert any(self.bm.logTransformParams) and not all(self.bm.logTransformParams)
        assert self.bm.useScaleFactors
        assert self.bm.parametersBounded
        lb = self.bm.lb
        ub = self.bm.ub
        # Check order of applying doesn't change bounds
        emptyMod = modification.Empty()
        emptyMod.apply(self.bm)
        mod = modification.Modification(logTransform='on', parameterBounds='on', scaleFactors='on')
        mod.apply_parameter_bounds(mod.dict['parameterBounds'], self.bm)
        mod.apply_log_transforms(mod.dict['log transform'], self.bm)
        mod.apply_scale_factors(mod.dict['scale factors'], self.bm)
        assert any(self.bm.logTransformParams) and not all(self.bm.logTransformParams)
        assert self.bm.useScaleFactors
        assert self.bm.parametersBounded
        assert all(np.array(self.bm.lb) == np.array(lb))
        assert all(np.array(self.bm.ub) == np.array(ub))

    def test_incorrect_settings(self):
        self.bm.reset()
        mod = modification.Modification(logTransform='not a setting', parameterBounds='no thank you', scaleFactors='yes please', rateBounds='i would rather not')
        # Should not change settings
        rb = self.bm.ratesBounded
        pb = self.bm.parametersBounded
        lt = self.bm.logTransformParams
        sft = self.bm.useScaleFactors
        mod.apply(self.bm)
        assert rb == self.bm.ratesBounded
        assert pb == self.bm.parametersBounded
        assert lt == self.bm.logTransformParams
        assert sft == self.bm.useScaleFactors

    @pytest.mark.cheap
    def test_staircase(self):
        # Check staircase problems override bounds
        mod = modification.Modification(logTransform='on', parameterBounds='off', scaleFactors='On', rateBounds='off')
        newbm = ionbench.problems.staircase.HH()
        mod.apply(newbm)
        assert newbm.useScaleFactors
        assert all(np.array(newbm.logTransformParams) == np.array(newbm.STANDARD_LOG_TRANSFORM))
        assert any(newbm.logTransformParams)
        assert not all(newbm.logTransformParams)
        assert newbm.parametersBounded
        assert newbm.ratesBounded


def test_loading():
    bm = ionbench.problems.test.Test()
    # Check all modifications load correctly
    for app in ionbench.APP_ALL:
        opt = app['module']
        module = import_module(opt)
        mod = module.get_modification(modNum=app['modNum'])
        mod.apply(bm)

    # Check empty modifications for any other modNum
    for opt in ionbench.OPT_ALL:
        module = import_module(opt)
        mod = module.get_modification(modNum=0)
        mod.apply(bm)



# Find the curved line between the steepest direction and the minimum is a quadratic shape is assumed based on the local curvature. Minimise along this line using Brent's method, as implemented in Press et al. "Numerical Recipes in C" by scipy. Update from Dokos 2004 compared with Dokos 2003 says that if you reach a local minimum whose cost is non-zero, you should weight the residuals and carry on

import numpy as np
import ionbench
import scipy


# noinspection PyShadowingNames
def run(bm, x0=None, maxIter=1000, maxInnerIter=100, costThreshold=0, debug=False):
    """
    Curvilinear gradient descent from Dokos 2004. This method was first introduced in Dokos 2003, with this implementation using the updated scheme which includes weighted residuals (Dokos 2004). It calculates a curved path which initially follows the steepest descent, but curves to finally converge towards a Gauss-Newton step (minimum of a locally defined quadratic). In order to optimise along this line, we use Scipy's implementation of Brent's method (an implementation based on Press et al. "Numerical Recipes in C").

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. If x0=None (the default), then the population will be sampled from the benchmarker problems .sample() method.
    maxIter : int, optional
        Maximum number of iterations of curvilinear gradient descent, summed up over all reweighting restarts. The default is 1000.
    maxInnerIter : int, optional
        Maximum number of iterations of curvilinear gradient descent before reweighting. The default is 100.
    costThreshold : float, optional
        The maximum acceptable weighted cost. Optimisation terminates only once this threshold is reached, or maxIter is exceeded. Default is 0.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """

    # Update to the weight vector for the residuals
    # noinspection PyShadowingNames
    def weight_update(wOld, r):
        """
        Update the weights for the residuals.
        Parameters
        ----------
        wOld : numpy array
            The previous weights.
        r : numpy array
            The residuals.

        Returns
        -------
        w : numpy array
            The updated weights.
        """
        rAbs = np.abs(r)
        mu1 = np.random.rand() * min(np.max(rAbs), 2 * np.median(rAbs))
        mu2 = np.random.rand() * min(np.max(rAbs), 2 * np.mean(rAbs))
        mu3 = (0.2 * np.random.rand() + 0.8) * np.max(rAbs)
        sigma = np.std(rAbs)
        dw = np.zeros(len(wOld))
        for i in range(len(dw)):
            dw[i] = np.exp(((rAbs[i] - mu1) / sigma) ** 2) + np.exp(((rAbs[i] - mu2) / sigma) ** 2) + np.exp(
                ((rAbs[i] - mu3) / sigma) ** 2)
        s = min(np.sqrt(10 * np.dot(wOld ** 2, r ** 2)), 1e5) / (np.sqrt(np.dot(dw ** 2, r ** 2)) + 1e-12)
        return wOld + s * dw

    # Set up transform functions to map between upper and lower bounds
    if not bm.parametersBounded:  # pragma: no cover
        bm.add_parameter_bounds()
    ub = bm.input_parameter_space(bm.ub)
    lb = bm.input_parameter_space(bm.lb)

    # noinspection PyShadowingNames
    def fitting_params(p):
        """
        Transform from model solve space to fitting space.
        Parameters
        ----------
        p : numpy array
            The parameters in model solve space.

        Returns
        -------
        x : numpy array
            The parameters in fitting space.
        """
        x = np.zeros(len(p))
        for i in range(len(x)):
            x[i] = np.arcsin(np.sqrt((p[i] - lb[i]) / (ub[i] - lb[i])))
        return x

    # noinspection PyShadowingNames
    def model_params(x):
        """
        Transform from fitting space to model solve space.
        Parameters
        ----------
        x : numpy array
            The parameters in fitting space.

        Returns
        -------
        p : numpy array
            The parameters in model solve space.
        """
        p = np.zeros(len(x))
        for i in range(len(p)):
            p[i] = lb[i] + (ub[i] - lb[i]) * np.sin(x[i]) ** 2
        return p

    signed_error = ionbench.utils.cache.get_cached_signed_error(bm)
    grad = ionbench.utils.cache.get_cached_grad(bm, returnCost=True, residuals=True)

    # sample initial point
    if x0 is None:
        x0 = bm.sample()
    # Transform to fitting space
    x0 = fitting_params(x0)

    # Initialise variables
    weightedSSE = 0  # Will store the weighted SSE from the start of a reweighting cycle
    setInitialWeights = True  # Ensures the initial weights wInit are set at the start and then not overridden
    reweightMaxIter = 10  # Maximum number of times to reweight (update the weights) before resetting them to their initial values
    weightCounter = 0  # Number of times the weights have been updated
    iterCounter = 0  # Total number of curvilinear gradient descent iterations, compared against maxIter every time a local minimum is found to check for termination
    bestParams = x0  # Best params found so far
    bestCost = np.inf  # Best unweighted cost found so far
    wInit = None  # Initial weights
    w = None  # Current weights
    while True:
        # Initialisation routine for every weight update/reset
        falphaOld = np.inf  # Stores the weighted SSE from the previous iteration in curvilinear GD, for termination criteria of curvilinear gradient descent
        # find residuals and jacobian. Residuals will be used to find weights, J will be used in first iteration of curvilinear GD
        r0, J = grad(model_params(x0))
        currentCost = np.sqrt(np.mean(r0 ** 2))
        if debug:
            print(f'Starting out optimisation loop. Weight counter {weightCounter}')

        # initial weights at the start of optimisation. Only run once.
        if setInitialWeights:
            wInit = 10 / (10 + np.abs(r0))
            setInitialWeights = False  # Ensure never run again.
            if debug:
                print('Setting initial weights')
                print('Initial weights set at:')
                print(wInit)
                print(f'Last residual was {np.abs(r0[-1])} which gave weight {wInit[-1]}')

        if weightCounter == 0:
            # If start of reweighting cycle, set initial weights
            w = wInit
            # Set initial weighted SSE
            weightedSSE = np.dot(w @ r0, w @ r0)
            if debug:
                print('Setting weights at initial values')
                print(f'Setting initial weighted SSE at {weightedSSE}')
        elif weightCounter == 1 and currentCost > bestCost:
            # Perturb the best parameters and make that the starting location. Perturbation chosen to match http://dx.doi.org/https://doi.org/10.26190/unsworks/21047
            if debug:
                print(
                    f'Current cost ({currentCost}) was actually worse than best so far ({bestCost}) so randomly perturbing parameters which gave best cost')
            x0 = bestParams * np.random.uniform(low=0.9, high=1.1, size=bm.n_parameters())
        elif np.dot(r0, r0) < 0.3 * weightedSSE:
            # Weighted SSE has decreased by 70% so reset weights
            if debug:
                print(
                    f'Weighted SSE significantly decreased. Original value was {weightedSSE} and it is now {np.dot(r0, r0)} after {weightCounter} new weights')
            weightCounter = 0
            continue  # Restart weights
        elif weightCounter >= reweightMaxIter:
            # Maximum number of reweights has been applied so reset weights
            if debug:
                print(
                    f'Maximum number of weight updates. Original weighted SSE was {weightedSSE} and it is now {np.dot(r0, r0)} after {weightCounter} new weights')
            weightCounter = 0
            continue  # Restart weights
        else:
            # Update the weights
            wTmp = np.copy(w)
            w = weight_update(w, r0)
            if debug:
                print(f'Applied weight update. First weight used to be {wTmp[0]}, and is now {w[0]}')

        if currentCost < bestCost:
            # Update which is the best point so far
            bestParams = x0
            bestCost = currentCost

        # Begin curvilinear gradient descent
        for k in range(maxInnerIter):
            # Update number of curvilinear GD iterations
            iterCounter += 1
            if debug:
                print(f'Curvilinear GD Iteration: {k} of {maxInnerIter}')
            if k > 0:
                # If first iteration of GD, r0 and J have already been calculated
                r0, J = grad(model_params(x0))

            # apply weights to residuals and jacobian
            r0 = np.diag(w) @ r0
            J = np.diag(w) @ J

            # transform jacobian into fitting space
            for i in range(bm.n_parameters()):
                J[:, i] = J[:, i] * (ub[i] - lb[i]) * 2 * np.sin(x0[i]) * np.cos(x0[i])

            # a=2*J^T*r0
            a = 2 * np.transpose(J) @ r0
            # H = 2*J^T*J
            H = 2 * np.transpose(J) @ J

            # Eigen value decomposition on hessian H
            try:  # pragma: no cover
                d, V = np.linalg.eig(H)
                invV = np.linalg.inv(V)
            except np.linalg.LinAlgError as e:
                print('Encountered a fatal error in optimisation. Infs or nans in Hessian. Terminating early')
                x0 = model_params(x0)
                bm.evaluate()
                return x0
            if debug:
                print('Eigenvalues:')
                print(d)

            # Generate function of the curved path from starting point (no step, L(0)=[0,0,0,...]), initially moving in the steepest descent direction, converging towards Gauss Newton step at alpha=inf
            # noinspection PyPep8Naming,PyShadowingNames
            def L(alpha):
                m = np.zeros(len(d))
                for i in range(len(d)):
                    if d[i] == 0:  # pragma: no cover
                        m[i] = -alpha
                    else:  # pragma: no cover
                        m[i] = (np.exp(-2 * d[i] * alpha) - 1) / (2 * d[i])
                M = np.diag(m)
                return V @ M @ invV @ a

            if debug:
                print(f'L(0): {L(0)}, L(inf): {L(np.inf)}')

            # Minimise bm.SSE(x0+L(alpha)) with respect to alpha using Brent's method
            # Define function to find weighted SSE from alpha
            # Cache SSE function to help make the code to handle the brent method more robust without needing repeated function evaluations
            # noinspection PyPep8Naming,PyShadowingNames
            def SSE(alpha):
                weightedr0 = np.diag(w) @ signed_error(model_params(x0 + L(alpha)))
                return np.dot(weightedr0, weightedr0)

            # Run Brent's method to optimise SSE with respect to alpha
            try:  # pragma: no cover
                out = scipy.optimize.brent(SSE, brack=(0, 1, 1e9), tol=1e-8, full_output=True)
                alpha = out[0]
                falphaNew = out[1]
            except ValueError as e:  # pragma: no cover
                # Failed. Time to find out why and try to recover
                # Option 1: NaNs in either SSE(0) or SSE(1e9) but not SSE(1) so step still viable
                if (np.isnan(SSE(0)) or np.isnan(SSE(1e9))) and not np.isnan(SSE(1)):
                    # If SSE(1) is NOT WORSE than SSE(0) or SSE(1e9) (accounting for SSE(0) or SSE(1e9) being NaN), then we can assume that it is a good step
                    if debug:
                        print(
                            f'NaNs were found. SSE(0): {SSE(0)}, SSE(1): {SSE(1)}, SSE(1e9): {SSE(1e9)}. Will try to continue if possible')
                    if not SSE(1) > SSE(0) and not SSE(1) > SSE(1e9):
                        alpha = 1
                        falphaNew = SSE(alpha)
                    elif not SSE(1e-6) > SSE(0) and not SSE(1e-6) > SSE(1e9):
                        alpha = 1e-6
                        falphaNew = SSE(alpha)
                    elif not SSE(1e-9) > SSE(0) and not SSE(1e-9) > SSE(1e9):
                        alpha = 1e-9
                        falphaNew = SSE(alpha)
                    else:
                        print(f'Failed. SSE(0): {SSE(0)}, SSE(1e-9): {SSE(1e-9)}, SSE(1e-6): {SSE(1e-6)}, SSE(1): {SSE(1)}, SSE(1e9): {SSE(1e9)}')
                        x0 = model_params(x0)
                        bm.evaluate()
                        return x0
                # Option 2: SSE(1e9) is best - take full step
                elif SSE(1e9) <= SSE(1) and SSE(1e9) <= SSE(0):
                    # SSE(1e9)<SSE(1),SSE(0)
                    alpha = 1e9
                    falphaNew = SSE(alpha)
                    if debug:
                        print('Alpha was minimised (at least approximately) by taking the full step')
                # Option 3: SSE(0) is smallest - start with even smaller step than 1
                elif SSE(0) <= SSE(1):
                    # SSE(0)<SSE(1),SSE(1e9)
                    if debug:
                        print(
                            f'SSE(0) {SSE(0)} is smaller than SSE(1) {SSE(1)}, trying smaller steps for the initial guesses (1e-6 and 1e-9). If not, will assume converged')
                    if SSE(1e-6) < SSE(0):
                        initStep = 1e-6
                    elif SSE(1e-9) < SSE(0):
                        initStep = 1e-9
                    else:
                        # The step needed would be too small so assume converged
                        x0 = model_params(x0)
                        bm.evaluate()
                        return x0
                    out = scipy.optimize.brent(SSE, brack=(0, initStep, 1e9), tol=1e-8, full_output=True)
                    alpha = out[0]
                    falphaNew = out[1]
                else:
                    print(f"Failed to find bounds for Brent's method. SSE(0): {SSE(0)}, SSE(1): {SSE(1)}, SSE(1e9): {SSE(1e9)}")
                    x0 = model_params(x0)
                    bm.evaluate()
                    return x0

            # Take new step
            x0 = x0 + L(alpha)

            if debug:
                print(f'Optimal alpha was {alpha}')
                print(f'Resulting step is {L(alpha)}')
                print(f'falphaOld: {falphaOld}, falphaNew: {falphaNew}')

            if falphaNew <= costThreshold:  # pragma: no cover
                # If cost is below threshold, fully terminate the optimisation
                if debug:
                    print(f'Cost below threshold {costThreshold} found. falphaOld: {falphaOld}, falphaNew: {falphaNew}')
                x0 = model_params(x0)
                bm.evaluate()
                return x0

            # Check for local minimum/slow convergence. From Dokos 2003.
            if np.abs(falphaNew - falphaOld) <= 0.0001 * falphaNew:
                if debug:
                    print(f'Local minimum found, {k} of {maxInnerIter}')
                break

            falphaOld = falphaNew

            if bm.is_converged():
                break

        weightCounter += 1
        if debug:
            print(f'Weight counter incremented to {weightCounter}')
        if iterCounter >= maxIter:  # pragma: no cover
            if debug:
                print(
                    f'Reached maximum number of iterations so terminating early. iterCounter: {iterCounter}, maxIter: {maxIter}')
            bm.set_max_iter_flag()
            break

        if bm.is_converged():
            break

    x0 = model_params(x0)
    bm.evaluate()
    return x0


# noinspection PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Dokos2004
    modNum = 2 -> Abed2013
    modNum = 3 -> Guo2010

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Dokos2004.

    """

    if modNum == 1:
        mod = ionbench.modification.Dokos2004()
    elif modNum == 2:
        mod = ionbench.modification.Abed2013()
    elif modNum == 3:
        mod = ionbench.modification.Guo2010()
    else:
        mod = ionbench.modification.Empty(name='dokos_gd')
    return mod



"""
A hybrid method of Nelder-Mead and PSO.
It uses a population of 3m+1 particles (where m = number of parameters) in which the best m+1 particles take a Nelder-Mead simplex step and the remaining 2m particles follow a PSO step.
The initial position sampling procedure is only partially defined. m+1 particles are generated like a NM simplex, but the remaining 2m particles are only defined as being generated randomly. We assume ionBench defaults for these particles.
No initial velocity sampling procedure is defined. We assume ionBench defaults.
For points generated by NM (accepting or shrinking), their velocity (in case they are demotes to the PSO step) is not defined. We assume a new randomly sampled velocity for new accepted points and the particles previous velocity for a simplex shrink step.
A particle's best position is treated similarly. Reset for accept and maintained for shrinking.
The PSO step is not defined. We assume a fairly typical setup with fixed acceleration coefficients at 1.496.
Guidance for the choice of eps is not given so a value of 1e-6 is used.
"""
import numpy as np
import ionbench


# noinspection PyShadowingNames
def run(bm, x0=None, maxIter=1000, eps=1e-6, debug=False):
    """
    Run the hybrid Nelder-Mead and PSO optimiser from Liu et al. 2011.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. Population is generated by randomly perturbing this initial guess +-50%, clamped to bounds if necessary. If x0=None (the default), then the population will be sampled uniformly between the bounds.
    maxIter : int, optional
        Maximum number of iterations. The default is 1000.
    eps : float
        Convergence criterion, roughly equivalent to a typical ftol criterion. The default is 0.01.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    if x0 is None:
        x0 = bm.sample()

    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    # noinspection PyShadowingNames
    class Particle(ionbench.utils.particle_optimisers.Particle):
        def __init__(self, position=None, transform=False):
            super().__init__(bm, cost_func, x0)
            if position is None:
                # Initial sampling (using util.particle_optimisers.Particle default)
                self.set_position(x0)
            else:
                # Force to be position (position is in input space, store in transformed [0,1] space)
                # This is only used for the initial population, after that the Nelder Mead and PSO are automatically using the transformed space.
                if transform:
                    self.position = self.transform(position)
                else:
                    self.position = np.copy(position)
            # Clamp to bounds
            self.clamp()

    # class for the simplex
    # noinspection PyShadowingNames
    class Simplex:
        def __init__(self, particles):
            self.particles = particles

        def get_best(self):
            # returns the best point in the simplex
            x_best = self.particles[0]
            for p in self.particles:
                if p.currentCost < x_best.currentCost:  # pragma: no cover
                    x_best = p
            return x_best

        def get_worst(self):
            # returns the worst point in the simplex
            x_worst = self.particles[0]
            for p in self.particles:
                if p.currentCost > x_worst.currentCost:
                    x_worst = p
            return x_worst

        def get_second_worst(self):
            # returns second-worst point in the simplex
            x_worst = self.get_worst()
            x_secondWorst = self.particles[0]
            for p in self.particles:
                if p.currentCost > x_secondWorst.currentCost and p != x_worst:
                    x_secondWorst = p
            return x_secondWorst

        def centroid(self, x_worst):
            # find the centroid of the simplex by averaging the positions, not including x_worst. Don't generate a point, cost not needed
            cen = [0] * bm.n_parameters()
            for p in self.particles:
                if p != x_worst:
                    cen += p.position
            cen /= len(self.particles) - 1
            return cen

        def accept(self, x):
            x_worst = self.get_worst()
            for i in range(len(self.particles)):
                if self.particles[i] == x_worst:
                    self.particles[i].position = x.position
                    self.particles[i].set_cost()
                    return

        def step(self):
            """
            Take a modified nelder mead step with acceptance probability related to the current temperature.
            """
            x_worst = self.get_worst()
            x_best = self.get_best()
            x_secondWorst = self.get_second_worst()
            c = self.centroid(x_worst)
            # attempt reflection
            xr = Particle(position=2 * c - x_worst.position)
            xr.set_cost()
            if x_secondWorst.currentCost > xr.currentCost >= x_best.currentCost:
                if debug:
                    print("Accept reflection")
                self.accept(xr)
                return
            if xr.currentCost < x_best.currentCost:
                # Attempt expand
                xe = Particle(position=2 * xr.position - c)
                xe.set_cost()
                if xe.currentCost < xr.currentCost:
                    if debug:
                        print("Accept expansion")
                    self.accept(xe)
                else:
                    if debug:
                        print("Ignore expansion. Accept reflection")
                    self.accept(xr)
                return
            if xr.currentCost >= x_secondWorst.currentCost:
                # contract
                if debug:
                    print("Attempt contraction")
                if x_worst.currentCost > xr.currentCost >= x_secondWorst.currentCost:
                    # outside contract
                    xc = Particle(position=(xr.position + c) / 2)
                    xc.set_cost()
                    if xc.currentCost <= xr.currentCost:
                        if debug:
                            print("Accept inside contraction")
                        self.accept(xc)
                        return
                else:
                    # inside contract
                    xc = Particle(position=(x_worst.position + c) / 2)
                    xc.set_cost()
                    if xc.currentCost < x_worst.currentCost:
                        if debug:
                            print("Accept inside contraction")
                        self.accept(xc)
                        return
            # shrink
            if debug:
                print("Shrink")
            for i in range(len(self.particles)):
                if self.particles[i] != x_best:
                    self.particles[i].position = (x_best.position + self.particles[i].position) / 2
                    self.particles[i].set_cost()
            return

        def stop(self):
            """
            Check if the simplex has converged.
            """
            costs = np.array([p.currentCost for p in self.particles])
            costs = costs[1:]
            Sc = np.sqrt(np.mean((np.mean(costs) - costs) ** 2))
            return Sc < eps

    L = 0
    assert bm.in_parameter_bounds(x0)
    assert bm.in_rate_bounds(x0)
    # Set population size n
    m = bm.n_parameters()
    n = 3 * m + 1
    # Initial population
    particleList = []
    for i in range(n):
        if i < m + 1:
            if i == 0:
                particleList.append(Particle(position=x0, transform=True))
            else:
                perturb = np.ones(m)
                perturb[i - 1] = 1.05
                particleList.append(Particle(position=x0 * perturb, transform=True))
        else:
            particleList.append(Particle())

    Gcost = [np.inf] * maxIter  # Best cost ever
    Gpos = [None] * maxIter  # Position of best cost ever
    for L in range(maxIter):
        if L > 0:
            Gcost[L] = Gcost[L - 1]
            Gpos[L] = Gpos[L - 1]

        if debug:
            print('-------------')
            print(f'Beginning population: {L}')
            print(f'Best cost so far: {Gcost[L]}')
            print(f'Found at position: {Gpos[L]}')

        for p in particleList:
            p.set_cost()
            if p.currentCost < Gcost[L]:
                Gcost[L] = p.currentCost
                Gpos[L] = np.copy(p.position)
        if L == 0:
            assert particleList[0].currentCost < 1e5
        # Sort particle list by cost
        costs = [p.currentCost for p in particleList]
        particleList = [p for _, p in sorted(zip(costs, particleList), key=lambda pair: pair[0])]

        # Build simplex and take 1 step
        s = Simplex(particleList[:m + 1])
        s.step()
        particleList[:m + 1] = s.particles

        # Renew velocities
        c1 = 1.496  # Assume they used the fixed value
        c2 = 1.496  # Assume they used the fixed value
        for p in particleList[m + 1:]:
            localAcc = c1 * np.random.rand() * (p.bestPosition - p.position)
            globalAcc = c2 * np.random.rand() * (Gpos[L] - p.position)
            p.velocity = p.velocity + localAcc + globalAcc
        if debug:
            print("Velocities renewed")
        # Move positions
        for p in particleList[m + 1:]:
            p.position += p.velocity
            # Enforce bounds by clamping
            p.clamp()

        if debug:
            print("Positions renewed")
            print(f'Finished population {L}')
            print(f'Best cost so far: {Gcost[L]}')
            print(f'Found at position: {Gpos[L]}')

        if bm.is_converged() or s.stop():
            break

    if L >= maxIter-1:
        bm.set_max_iter_flag()

    bm.evaluate()
    return Particle().untransform(Gpos[L])


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Liu2011

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Liu2011.

    """
    mod = ionbench.modification.Liu2011()
    return mod



import numpy as np
import copy

from pymoo.core.individual import Individual as pymooInd
from pymoo.core.population import Population
from pymoo.core.problem import Problem
from pymoo.operators.crossover.sbx import SBX
from pymoo.operators.crossover.pntx import SinglePointCrossover
from pymoo.operators.mutation.pm import PolynomialMutation


class Individual:
    """
    Individual class for genetic algorithms.
    """

    def __init__(self, bm, x0, cost_func):
        """
        Initialise an individual.
        Parameters
        ----------
        bm : ionbench.Benchmarker
            Benchmarker object.
        x0 : np.ndarray
            Initial guess for the population. Individual will be sampled around x0.
        cost_func : function
            Cost function.
        """
        if x0 is None:
            self.x = bm.sample()
        else:
            self.x = bm.input_parameter_space(
                bm.original_parameter_space(x0) * np.random.uniform(low=0.5, high=1.5, size=bm.n_parameters()))
        self.x = bm.clamp_parameters(self.x)
        self.cost = None
        self.cost_func = cost_func

    def find_cost(self):
        """
        Find the cost of the individual for its current self.x. Updates self.cost.
        """
        self.cost = self.cost_func(tuple(self.x))


def get_pop(bm, x0, n, cost_func):
    """
    Generate a population of n individuals.

    Parameters
    ----------
    bm : ionbench.Benchmarker
        Benchmarker object.
    x0 : np.ndarray
        Initial guess for the population to generate around.
    n : int
        Number of individuals to generate.
    cost_func : function
        Cost function.

    Returns
    -------
    pop : list
        List of individuals.
    """
    pop = [None] * n
    for i in range(n):
        pop[i] = Individual(bm, x0, cost_func)
    return find_pop_costs(pop)


def find_pop_costs(pop):
    """
    Find the costs of the population by applying pop[i].find_cost() to all individuals.

    Parameters
    ----------
    pop : list
        List of individuals.

    Returns
    -------
    pop : list
        List of individuals with costs updated.
    """
    for i in range(len(pop)):
        pop[i].find_cost()
    return pop


def tournament_selection(pop):
    """
    Perform tournament selection on the population.

    Parameters
    ----------
    pop : list
        List of individuals.

    Returns
    -------
    newPop : list
        New population after tournament selection.
    """
    newPop = []
    for j in range(2):
        perm = np.random.permutation(len(pop))
        for i in range(len(pop) // 2):
            if pop[perm[2 * i]].cost < pop[perm[2 * i + 1]].cost:
                newPop.append(copy.deepcopy(pop[perm[2 * i]]))
            else:
                newPop.append(copy.deepcopy(pop[perm[2 * i + 1]]))
    return newPop  # Population of parents


def one_point_crossover(pop, bm, cost_func, crossoverProb=0.5):
    """
    Perform one point crossover on the population.

    Parameters
    ----------
    pop : list
        List of individuals
    bm : ionbench.Benchmarker
        Benchmarker object
    cost_func : function
        Cost function
    crossoverProb : float, optional
        Probability of crossover. The default is 0.5.

    Returns
    -------
    newPop : list
        New population after crossover.
    """
    newPop = []
    problem = Problem(n_var=bm.n_parameters(), xl=bm.input_parameter_space(bm.lb), xu=bm.input_parameter_space(bm.ub))
    for i in range(len(pop) // 2):
        a, b = pymooInd(X=np.array(pop[2 * i].x)), pymooInd(X=np.array(pop[2 * i + 1].x))
        parents = [[a, b]]
        off = SinglePointCrossover(prob=crossoverProb).do(problem, parents)
        newPop = add_pymoo(bm, newPop, off, cost_func)
    return newPop


def add_pymoo(bm, pop, off, cost_func):
    """
    Add the pymoo population to the ionbench population list.
    Parameters
    ----------
    bm : ionbench.Benchmarker
        Benchmarker object
    pop : list
        List of individuals (ionbench population)
    off : pymoo.core.population.Population
        Population of individuals (pymoo population)
    cost_func : function
        A cost function to use for the ionbench individuals.

    Returns
    -------
    pop : list
        List of individuals (ionbench population) with the new individuals added.
    """
    Xp = off.get("X")
    for i in range(len(Xp)):
        pop.append(Individual(bm, bm.sample(), cost_func))
        pop[-1].x = Xp[i]
    return pop


def sbx_crossover(pop, bm, cost_func, eta_cross):
    """
    Perform one point crossover on the population.

    Parameters
    ----------
    pop : list
        List of individuals
    bm : ionbench.Benchmarker
        Benchmarker object
    cost_func : function
        Cost function
    eta_cross : float
        Parameter for the crossover.

    Returns
    -------
    newPop : list
        New population after crossover.
    """
    newPop = []
    problem = Problem(n_var=bm.n_parameters(), xl=bm.input_parameter_space(bm.lb), xu=bm.input_parameter_space(bm.ub))
    for i in range(len(pop) // 2):
        a, b = pymooInd(X=np.array(pop[2 * i].x)), pymooInd(X=np.array(pop[2 * i + 1].x))
        parents = [[a, b]]
        off = SBX(prob=0.9, prob_var=0.5, eta=eta_cross).do(problem, parents)
        newPop = add_pymoo(bm, newPop, off, cost_func)
    return newPop


def polynomial_mutation(pop, bm, cost_func, eta_mut):
    """
    Perform polynomial mutation from Bot et al. 2012 on the population.

    Parameters
    ----------
    pop : list
        List of individuals
    bm : ionbench.Benchmarker
        Benchmarker object
    cost_func : function
        Cost function
    eta_mut : float
        Parameter for the mutation.

    Returns
    -------
    newPop : list
        New population after crossover.
    """
    problem = Problem(n_var=bm.n_parameters(), xl=bm.input_parameter_space(bm.lb), xu=bm.input_parameter_space(bm.ub))
    mutation = PolynomialMutation(prob=1, prob_var=0.1, eta=eta_mut)
    for i in range(len(pop)):
        ind = Population.new(X=[pop[i].x])
        off = mutation(problem, ind)
        pop[i] = Individual(bm, bm.sample(), cost_func)
        pop[i].x = off.get("X")[0]
    return pop


def get_elites(pop, n):
    """
    Get the n elite individuals from the population.

    Parameters
    ----------
    pop : list
        List of individuals
    n : int
        Number of elite individuals to return

    Returns
    -------
    list
        List of elite individuals
    """
    costVec = [pop[i].cost for i in range(len(pop))]
    eliteIndices = np.argsort(costVec)[:n]
    elites = [copy.deepcopy(pop[eliteIndices[i]]) for i in range(n)]
    return elites


def set_elites(pop, elites):
    """
    Set the elite individuals in the population.

    Parameters
    ----------
    pop : list
        List of individuals.
    elites : list
        List of elite individuals.

    Returns
    -------
    pop : list
        Copy of the list of individuals (pop) with the elite individuals replacing the worst individuals.
    """
    costVec = [pop[i].cost for i in range(len(pop))]
    eliteIndices = np.argsort(costVec)[-len(elites):]
    for i in range(len(elites)):
        pop[eliteIndices[i]] = copy.deepcopy(elites[i])
    return copy.deepcopy(pop)



import numpy as np
import ionbench


# noinspection PyShadowingNames
def run(bm, x0=None, maxIter=1000, debug=False):
    """
    Random search from Vanier et al. 1999. This optimiser randomly samples maxIter parameters and reports the best one.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. If x0=None (the default), then the population will be sampled from the benchmarker problems .sample() method.
    maxIter : int, optional
        Maximum number of iterations. The default is 1000.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters found.

    """
    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    if x0 is None:
        # sample initial point
        x0 = bm.sample()
    x_best = x0
    cost_best = cost_func(x0)
    if debug:
        print(f'Starting cost is {cost_best}')
    i = None
    for i in range(maxIter):
        x_new = bm.sample()
        cost_new = cost_func(x_new)
        if debug:
            print(f'New point has cost of {cost_new}')
        if cost_new < cost_best:
            cost_best = cost_new
            x_best = x_new
            if debug:
                print('Improvement found')
        if bm.is_converged():
            break

    if i >= maxIter-1:
        bm.set_max_iter_flag()

    if debug:
        print('Complete')
        print(f'Final cost is {cost_best}')

    bm.evaluate()
    return x_best


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Vanier1999

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Vanier1999.

    """
    mod = ionbench.modification.Vanier1999()
    return mod



"""
Contains the test benchmarker problem. This inherits from benchmarker.Benchmarker but with many of the methods overwritten. It is used to test the optimisers are able to sufficiently optimise on a problem that is much quicker to solve.
generate_data() will generate the data for test problem and store it in the data directory.
"""
import os
import numpy as np
import csv
import ionbench
from scipy.stats import norm
import warnings
import time

import ionbench.tracker.tracker


class Test(ionbench.benchmarker.Benchmarker):
    def __init__(self):
        self.NAME = "test"
        self.COST_THRESHOLD = 0.001
        self._TRUE_PARAMETERS = np.array([2., 4.])
        self._RATE_FUNCTIONS = ()
        self.RATE_MIN = 0
        self.RATE_MAX = 1
        self.STANDARD_LOG_TRANSFORM = (False, True)
        self.sensitivityCalc = True
        self.T_MAX = 20
        self.TIMESTEP = 1  # Timestep in data between points
        try:
            self.load_data(os.path.join(ionbench.DATA_DIR, 'test', 'data.csv'))
        except FileNotFoundError:  # pragma: no cover
            self.DATA = None
        self._LOWER_BOUND = self._TRUE_PARAMETERS * 0.5
        self._UPPER_BOUND = self._TRUE_PARAMETERS * 1.5
        self.lb = np.copy(self._LOWER_BOUND)
        self.ub = np.copy(self._UPPER_BOUND)
        self.sim = FakeSim()
        self.simSens = FakeSim()
        super().__init__()

    def sample(self, n=1):
        """
        Sample parameters for the test problem.

        Parameters
        ----------
        n : int, optional
            Number of parameter vectors to sample. The default is 1.

        Returns
        -------
        params : list
            If n=1, then params is the vector of parameters. Otherwise, params is a list containing n parameter vectors.

        """
        params = [None] * n
        for i in range(n):
            params[i] = self.input_parameter_space(
                self._TRUE_PARAMETERS * np.random.uniform(0.5, 1.5, self.n_parameters()))
        if n == 1:
            return params[0]
        else:
            return params

    def solve_model(self, times, continueOnError=True):
        out = norm(self.sim.parameters['.p1'], self.sim.parameters['.p2']).pdf(times)
        return out

    def solve_with_sensitivities(self, times):
        parameters = np.array([self.simSens.parameters['.p1'], self.simSens.parameters['.p2']])
        curr = norm(self.sim.parameters['.p1'], self.sim.parameters['.p2']).pdf(times)
        sens = np.zeros((len(times), 1, self.n_parameters()))
        for t in range(len(times)):
            sens[t, 0, 0] = curr[t] * (t - parameters[0]) / parameters[1] ** 2
            sens[t, 0, 1] = curr[t] * ((t - parameters[0]) ** 2 / parameters[1] ** 3 - 1 / parameters[1])
        return curr, sens

    def use_sensitivities(self):  # pragma: no cover
        # Not needed for test function. Override to avoid trying to access myokit objects
        pass

    def evaluate(self):  # pragma: no cover
        # Not needed for test function. Override to avoid trying to access myokit objects
        pass

    def set_steady_state(self, parameters):
        pass


class FakeSim:
    """
    A fake myokit simulation object to ignore any Simulation methods for the test problem.
    It stores set parameters.
    """
    def __init__(self):
        self.parameters = {}

    def reset(self):
        pass

    def set_constant(self, name, value):
        self.parameters[name] = value
        pass


# noinspection PyProtectedMember
def generate_data():  # pragma: no cover
    """
    Generate the data files for the test benchmarker problem.

    Parameters
    ----------
    None.

    Returns
    -------
    None.

    """
    bm = Test()
    out = bm.simulate(bm._TRUE_PARAMETERS, np.arange(0, bm.T_MAX, bm.TIMESTEP), continueOnError=False)
    with open(os.path.join(ionbench.DATA_DIR, 'test', 'data.csv'), 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',')
        writer.writerows(map(lambda x: [x], out))



"""
Contains the INa benchmark problem based on the problem from Moreno et al. 2016, which some changes to the voltage clamp protocol for speeding up model evaluation. This inherits from benchmarker.Benchmarker.
generate_data() will generate the data the INa problem and store it in the data directory.
"""
import ionbench
import myokit
import os
import numpy as np
import csv
import warnings
from myokit.lib.markov import LinearModel


class INa(ionbench.benchmarker.Benchmarker):
    """
    The Moreno 2016 INa benchmarker.

    The benchmarker uses the model from Moreno et al. 2016 with a step protocol used to calculated summary curves which are then used for fitting.

    Its parameters are specified as reported in Moreno et al. 2016 with the true parameters being the same as the default and the center of the sampling distribution.
    """

    def __init__(self, sensitivities=False):
        print('Initialising Moreno 2016 INa benchmark')
        # Benchmarker
        self.NAME = "moreno2016.ina"
        self._TRUE_PARAMETERS = np.array(
            [7.6178e-3, 3.2764e1, 5.8871e-1, 1.5422e-1, 2.5898, 8.5072, 1.3760e-3, 2.888, 3.2459e-5, 9.5951, 1.3771,
             2.1126e1, 1.1086e1, 4.3725e1, 4.1476e-2, 2.0802e-2])
        self.STANDARD_LOG_TRANSFORM = (True, False, True, True) * 2 + (True, False) * 3 + (True,) * 2
        self._RATE_FUNCTIONS = (lambda p, V: 1 / (p[0] * np.exp(-V / p[1])),
                                lambda p, V: p[2] / (p[0] * np.exp(-V / p[1])),
                                lambda p, V: p[3] / (p[0] * np.exp(-V / p[1])),
                                lambda p, V: 1 / (p[4] * np.exp(V / p[5])),
                                lambda p, V: p[6] / (p[4] * np.exp(V / p[5])),
                                lambda p, V: p[7] / (p[4] * np.exp(V / p[5])),
                                lambda p, V: p[8] * np.exp(-V / p[9]),
                                lambda p, V: p[10] * np.exp(V / p[11]),
                                lambda p, V: p[12] * np.exp(V / p[13]),
                                lambda p, V: p[3] / (p[0] * np.exp(-V / p[1])) * p[12] * np.exp(V / p[13]) * p[8] * np.exp(-V / p[9]) / (
                                                      p[7] / (p[4] * np.exp(V / p[5])) * p[10] * np.exp(V / p[11])),
                                lambda p, V: p[14] * p[12] * np.exp(V / p[13]),
                                lambda p, V: p[15] * p[8] * np.exp(-V / p[9]))  # Used for rate bounds
        self.sensitivityCalc = sensitivities
        self.COST_THRESHOLD = 1.70e-6
        protocol = self.protocol()
        self.TIMESTEP = 0.5  # Timestep in data between points
        self.T_MAX = protocol.characteristic_time()
        self._logTimes = None
        self._ssiBounds = None
        self._actBounds = None
        self._rfiBounds = None
        self._tauBounds = None
        self._LOWER_BOUND = self._TRUE_PARAMETERS * 0.75
        self._UPPER_BOUND = self._TRUE_PARAMETERS * 1.25
        self.lb = np.copy(self._LOWER_BOUND)
        self.ub = np.copy(self._UPPER_BOUND)
        self.RATE_MIN = 1.67e-5
        self.RATE_MAX = 1e3
        self.load_data(dataPath=os.path.join(ionbench.DATA_DIR, 'moreno2016', 'ina.csv'))
        w = np.array([1 / 9] * 9 + [1 / 20] * 20 + [1 / 10] * 10 + [1 / 9] * 9)
        self.WEIGHTS = w / np.sum(w)

        # Myokit
        self._MODEL = myokit.load_model(os.path.join(ionbench.DATA_DIR, 'moreno2016', 'moreno2016.mmt'))
        self._OUTPUT_NAME = 'ina.INa'
        self._PARAMETER_CONTAINER = 'ina'
        parameters = [self._PARAMETER_CONTAINER + '.p' + str(i + 1) for i in range(self.n_parameters())]
        self._ANALYTICAL_MODEL = LinearModel(model=self._MODEL, parameters=parameters, current=self._OUTPUT_NAME,
                                             states=['ina.' + s for s in ['ic3', 'ic2', 'if', 'c3', 'c2', 'c1', 'o',
                                                                          'is']], vm='membrane.V')
        self._TOLERANCES = (1e-7, 1e-7)
        self.sim = myokit.lib.markov.AnalyticalSimulation(self._ANALYTICAL_MODEL, protocol=self.protocol())
        if self.sensitivityCalc:
            # ODE solver
            sens = ([self._OUTPUT_NAME], parameters)
            self.simSens = myokit.Simulation(self._MODEL, sensitivities=sens, protocol=self.protocol())
            self.simSens.set_tolerance(*self._TOLERANCES)
        else:
            self.simSens = None
        super().__init__()
        print('Benchmarker initialised')

    def sample(self, n=1):
        """
        Sample parameters for the Moreno 2016 problems.

        Parameters
        ----------
        n : int, optional
            Number of parameter vectors to sample. The default is 1.

        Returns
        -------
        params : list
            If n=1, then params is the vector of parameters. Otherwise, params is a list containing n parameter vectors.

        """
        params = [None] * n
        for i in range(n):
            param = [None] * self.n_parameters()
            for j in range(self.n_parameters()):
                param[j] = self._TRUE_PARAMETERS[j] * np.random.uniform(0.75, 1.25)
            params[i] = self.input_parameter_space(param)
        if n == 1:
            return params[0]
        else:
            return params

    def protocol(self):
        """
        Add the protocol from Moreno et al. 2016. This is a series of steps, used for forming the summary statistic curves. This also sets the time points that need to be recorded to evaluate these summary statistics.

        Returns
        -------
        p : myokit.Protocol
            A myokit.Protocol for the voltage clamp protocol from Moreno et al. 2016.

        """
        # Setup
        measurementWindows = []
        gap = 5000
        newProtocol = myokit.Protocol()

        # Protocol 1 - Measure peak current at -10mV (25ms) after holding at voltages between -120mV and -40mV (5000ms)
        # Track start times
        protocolStartTimes = [0]
        # Create protocol
        protocol = myokit.pacing.steptrain_linear(vstart=-120, vend=-30, dv=10, vhold=-10, tpre=0, tstep=gap, tpost=25)
        # Add windows to measure at
        for e in protocol.events():
            if e.level() == -10:
                measurementWindows.append([e.start(), e.stop(), 'ssi'])
        # Add protocol to full protocol
        for e in protocol.events():
            newProtocol.schedule(level=e.level(), start=e.start(), duration=e.duration())
        # Add barrier to separate effects from different protocols
        newProtocol.schedule(level=-100, start=newProtocol.characteristic_time(), duration=gap)

        # Protocol 2 - Measure steady state current at varying voltages between -75mV and 20mV
        # Track start times
        protocolStartTimes.append(newProtocol.characteristic_time())
        # Create protocol
        protocol = myokit.pacing.steptrain_linear(vstart=-75, vend=25, dv=5, vhold=-100, tpre=gap, tstep=gap, tpost=0)
        # Add windows to measure at
        offset = newProtocol.characteristic_time()
        for e in protocol.events():
            if e.level() != -100:
                measurementWindows.append([e.stop() + offset - 0.01, 'act'])  # Just before transition
        # Add protocol to full protocol
        for e in protocol.events():
            newProtocol.schedule(level=e.level(), start=offset + e.start(), duration=e.duration())
        # Add barrier to separate effects from different protocols
        newProtocol.schedule(level=-100, start=newProtocol.characteristic_time(), duration=gap)

        # Protocol 3 - Ratio of max current at steps to -10mV with varying length gaps between
        # Track start times
        protocolStartTimes.append(newProtocol.characteristic_time())
        # Create protocol
        # dt = [1, 5, 10, 25, 50, 100, 150, 250, 500, 1000, 2000, 3000, 5000, 7500, 10000]
        dt = [1, 5, 10, 25, 50, 100, 150, 250, 500,
              1000]  # Cut short since larger dt all round to 1 in data so remaining points will be most informative
        vsteps = [-100, -10, -100, -10] * len(dt) + [-100]
        times = []
        for i in range(len(dt)):
            times.append(gap)
            times.append(100)
            times.append(dt[i])
            times.append(25)
        times.append(gap)
        protocol = myokit.Protocol()
        for i in range(len(times)):
            protocol.add_step(vsteps[i], times[i])
        # Add windows to measure at
        offset = newProtocol.characteristic_time()
        for e in protocol.events():
            if e.level() == -10:
                measurementWindows.append([e.start() + offset, e.stop() + offset, 'rfi'])
        # Add protocol to full protocol
        for e in protocol.events():
            newProtocol.schedule(level=e.level(), start=offset + e.start(), duration=e.duration())
        # Add barrier to separate effects from different protocols
        newProtocol.schedule(level=-100, start=newProtocol.characteristic_time(), duration=gap)

        # Protocol 5 - Time to 50% of max current after step to between -20mV and 20mV
        # Track start times
        protocolStartTimes.append(newProtocol.characteristic_time())
        # Create protocol
        protocol = myokit.pacing.steptrain_linear(vstart=-20, vend=25, dv=5, vhold=-100, tpre=gap, tstep=40, tpost=0)
        # Add windows to measure at
        offset = newProtocol.characteristic_time()
        for e in protocol.events():
            if e.level() != -100:
                measurementWindows.append([e.start() + offset, e.stop() + offset, 'tau'])
        # Add protocol to full protocol
        for e in protocol.events():
            newProtocol.schedule(level=e.level(), start=offset + e.start(), duration=e.duration())

        # Track total time
        protocolStartTimes.append(newProtocol.characteristic_time())

        # Store measurement windows
        self._logTimes = []
        self._ssiBounds = []
        self._actBounds = []
        self._rfiBounds = []
        self._tauBounds = []

        for i in measurementWindows:
            if i[-1] == 'act':
                self._logTimes.append(i[0])
                self._actBounds.append(len(self._logTimes) - 1)
            else:
                lb = len(self._logTimes)
                self._logTimes += list(np.arange(i[0], i[1], 0.005))
                ub = len(self._logTimes)
                if i[-1] == 'ssi':
                    self._ssiBounds.append([lb, ub])
                elif i[-1] == 'rfi':
                    self._rfiBounds.append([lb, ub])
                elif i[-1] == 'tau':
                    self._tauBounds.append([lb, ub])
        self._logTimes = np.array(self._logTimes)
        return newProtocol

    def solve_with_sensitivities(self, times):
        """
        Overrides Benchmarker.solve_with_sensitivities() in order to calculate sensitivities of summary statistics rather than current.
        Parameters
        ----------
        times : list
            Unused. Would be used to specify a list of times at which to record the model but the Moreno problem does not use that. Only kept here to ensure compatibility with the other problems.

        Returns
        -------
        curr : numpy array
            The summary statistics of the solved current.
        sens : numpy array
            The summary statistic sensitivities of the solved current.
        """
        log, sens = self.simSens.run(self.T_MAX + 1, log_times=self._logTimes)
        curr = np.array(log[self._OUTPUT_NAME])
        sens = np.array(sens)
        # Adjust sens to emulate moreno summary statistics
        sens_SS = np.zeros((len(self.DATA), 1, self.n_parameters()))
        for i in range(self.n_parameters()):
            step = 1e-5 * self._TRUE_PARAMETERS[i]
            sens_SS[:, 0, i] = (self.sum_stats(curr + step * sens[:, 0, i]) - self.sum_stats(
                curr - step * sens[:, 0, i])) / (2 * step)
        curr = self.sum_stats(curr)
        sens = sens_SS
        return curr, sens

    def solve_model(self, times, continueOnError=True):
        """
        Replaces the Benchmarker solve_model to call a special Moreno 2016 method (sum_stats()) which handles the summary curve calculations. The output is a vector of points on the summary curves.

        Parameters
        ----------
        times : list or numpy array
            Unnecessary for Moreno 2016. Only kept in since it will be passed in as an input by the main Benchmarker methods.
        continueOnError : bool, optional
            If continueOnError is True, any errors that occur during solving the model will be ignored and an infinite output will be given. The default is True.

        Returns
        -------
        modelOutput : list
            A vector of points on summary curves.

        """
        if continueOnError:
            try:
                # Run a simulation
                log = self.sim.run(self.T_MAX + 1, log_times=self._logTimes)
                # log = self.sim.run(self.T_MAX + 1, log_times=self._logTimes, log=[self._OUTPUT_NAME]) # Setting output name only works for ODE sims, not analytical
                return self.sum_stats(np.array(log[self._OUTPUT_NAME], dtype='float64'))
            except (myokit.SimulationError, np.linalg.LinAlgError):  # pragma: no cover
                warnings.warn("Failed to solve model. Will report infinite output in the hope of continuing the run.")
                return np.array([np.inf] * len(self.DATA), dtype='float64')
        else:
            log = self.sim.run(self.T_MAX + 1, log_times=self._logTimes)
            # log = self.sim.run(self.T_MAX + 1, log_times=self._logTimes, log=[self._OUTPUT_NAME]) # Setting outputName only works for ODE sims, not analytical
            return self.sum_stats(np.array(log[self._OUTPUT_NAME], dtype='float64'))

    def rmse(self, c1, c2):
        """
        Returns the RMSE between c1 and c2. This function is overrides the general benchmarker function to do weighted RMSE.

        Parameters
        ----------
        c1 : numpy array
            A list of model outputs, typically current.
        c2 : numpy array
            The data to compare the model output to. Should be the same size as c1.

        Returns
        -------
        rmse : float
            The weighted RMSE between c1 and c2.
        """
        return np.sqrt(np.average((c1 - c2) ** 2, weights=self.WEIGHTS))

    def sum_stats(self, inaOut):
        """
        Runs the model to generate the Moreno et al. 2016 summary curves. The points on these summary curves are then returned.
        
        Parameters
        ----------
        inaOut : numpy array
            The iNa current trace on which to calculate the summary statistics.
        Returns
        -------
        modelOutput : list
            A list of points on summary curves.

        """
        # Flip sign of current for max current calculations
        inaOut = -inaOut

        ssi = [-1] * 9
        for i in range(len(ssi)):
            ssi[i] = max(inaOut[self._ssiBounds[i][0]:self._ssiBounds[i][1]])
        ssi = [a / max(ssi) for a in ssi]

        act = [-1] * 20
        for i in range(len(act)):
            act[i] = inaOut[self._actBounds[i]]
        act = [a / max(act) for a in act]

        rfi = [-1] * 10
        for i in range(len(rfi)):
            firstPulse = max(inaOut[self._rfiBounds[2 * i][0]:self._rfiBounds[2 * i][1]])
            secondPulse = max(inaOut[self._rfiBounds[2 * i + 1][0]:self._rfiBounds[2 * i + 1][1]])
            rfi[i] = secondPulse / firstPulse

        tau = [-1] * 9
        for i in range(len(tau)):
            peak = max(inaOut[self._tauBounds[i][0]:self._tauBounds[i][1]])
            for j in range(self._tauBounds[i][1] - self._tauBounds[i][0]):
                current = inaOut[self._tauBounds[i][0] + j]
                if current == peak:
                    timeToPeak = self._logTimes[self._tauBounds[i][0] + j]
                    tau[i] = np.interp(np.abs(peak / 2),
                                       np.abs(np.flip(inaOut[self._tauBounds[i][0] + j:self._tauBounds[i][1]])),
                                       np.flip(self._logTimes[
                                               self._tauBounds[i][0] + j:self._tauBounds[i][1]])) - timeToPeak
                    break
        return np.array(ssi + act + rfi + tau)


# noinspection PyProtectedMember
def generate_data():  # pragma: no cover
    """
    Generate the data files for the Moreno 2016 benchmarker problems. The true parameters are the same as the default for this benchmark problem.

    Parameters
    ----------
    None.

    Returns
    -------
    None.

    """
    bm = INa()
    bm.set_params(bm._TRUE_PARAMETERS)
    bm.set_steady_state(bm._TRUE_PARAMETERS)
    out = bm.solve_model(np.arange(0, bm.T_MAX, bm.TIMESTEP), continueOnError=False)
    with open(os.path.join(ionbench.DATA_DIR, 'moreno2016', 'ina.csv'), 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',')
        writer.writerows(map(lambda x: [x], out))



"""
This module contains the conjugate gradient descent method implemented using scipy.optimize.minimize.
It is an unconstrained optimiser and will struggle with the staircase problems due to the discontinuous nature of the cost/penalty function.
"""
import ionbench.problems.staircase
import scipy.optimize


# noinspection PyShadowingNames
def run(bm, x0=None, gtol=0.001, maxIter=1000, debug=False):
    """
    Runs Conjugate Gradient Descent optimiser from Scipy.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is None, in which case a randomly sampled parameter vector is retrieved from bm.sample().
    gtol : float, optional
        Tolerance in for the gradient. Gradient norm must be less than gtol before algorithm successfully terminates. The default is 0.001.
    maxIter : int, optional
        Maximum number of iterations of Conjugate Gradient Descent to use. The default is 1000.
    debug : bool, optional
        If True, prints out the cost and parameters found by the algorithm. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified by Conjugate Gradient Descent.
    """
    if x0 is None:
        x0 = bm.sample()
        if debug:
            print('Sampling x0')
            print(x0)

    out = scipy.optimize.minimize(bm.cost, x0, jac=bm.grad, method='CG', options={'disp': debug, 'gtol': gtol, 'maxiter': maxIter})

    if out.nit >= maxIter:
        bm.set_max_iter_flag()

    if debug:
        print(f'Cost of {out.fun} found at:')
        print(out.x)

    bm.evaluate()
    return out.x


# noinspection PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Sachse2003
    modNum = 2 -> Vanier1999

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Sachse2003.

    """

    if modNum == 1:
        mod = ionbench.modification.Sachse2003()
    elif modNum == 2:
        mod = ionbench.modification.Vanier1999()
    else:
        mod = ionbench.modification.Empty(name='conjugateGD_scipy_scipy')
    return mod



"""
Generates and plots profile likelihoods of benchmarker problems. Contains the class ProfileManager which masks a benchmarker but has a reduced parameter size. Contains the functions run(), which generates and saves the profile likelihoods, and plot_profile_likelihood(), which plots the saved results.
"""
import ionbench
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import pickle
import os
import scipy


class ProfileManager:
    """
    A middle-man for handling profile likelihood plots. This class acts as a wrapper for the benchmarker, such that the benchmarker always sees the full parameter vector, while the optimiser sees a reduced vector with one parameter missing.
    Contains methods for n_parameters(), cost(), grad(), signed_error(), set_params(), sample(), and evaluate(), all of which handle the mapping between the full set of parameters and the reduced optimised set of parameters.
    """

    def __init__(self, bm, fixedParam, fixedValue, x0):
        """
        Construct a ProfileManager.

        Parameters
        ----------
        bm : benchmarker
            A benchmarker to act as the wrapper for.
        fixedParam : int
            The index of the parameter to be fixed.
        fixedValue : float
            The value at which to fix the parameter.
        x0 : list
            Initial parameter vector (including the fixed parameter) to begin optimisation

        Returns
        -------
        None.

        """
        self.bm = bm
        self.fixedParam = fixedParam
        self.fixedValue = fixedValue
        self.MLE = x0
        self.parametersBounded = False
        self.NAME = bm.NAME
        self.T_MAX = bm.T_MAX
        self.TIMESTEP = bm.TIMESTEP
        self.DATA = bm.DATA

    def n_parameters(self):  # pragma: no cover
        """
        ProfileManager wrapper for the benchmarker's .n_parameters() method. Reduces the parameter count by 1 to account for the fixed parameter.
        """
        return self.bm.n_parameters() - 1

    def cost(self, params):  # pragma: no cover
        """
        ProfileManager wrapper for the benchmarker's .cost() method. Inserts the fixed parameter before evaluating.
        """
        return self.bm.cost(self.set_params(params))

    def grad(self, params, **kwargs):  # pragma: no cover
        """
        ProfileManager wrapper for the benchmarker's .grad() method. 
        """
        return np.delete(self.bm.grad(self.set_params(params), **kwargs), self.fixedParam, 1)

    def signed_error(self, params):  # pragma: no cover
        """
        ProfileManager wrapper for the benchmarker's .signed_error() method. Inserts the fixed parameter before evaluating.
        """
        return self.bm.signed_error(self.set_params(params))

    def set_params(self, params):  # pragma: no cover
        """
        Inserts the fixed parameter into the inputted parameter vector.
        """
        return np.insert(params, self.fixedParam, self.fixedValue)

    def sample(self):  # pragma: no cover
        """
        ProfileManager wrapper for the benchmarker's .sample() method. Since sampling is used to define the initial points, this method returns the true parameters with the fixed parameter removed.
        """
        return np.delete(self.MLE, self.fixedParam)

    def evaluate(self):  # pragma: no cover
        """
        ProfileManager wrapper for the benchmarker's .evaluate() method. This method does nothing, but is necessary since optimisers will attempt to call it.
        """
        pass

    def set_max_iter_flag(self):  # pragma: no cover
        """
        ProfileManager wrapper for the benchmarker's .set_max_iter_flag() method. This method does nothing, but is necessary since optimisers will attempt to call it.
        """
        pass


# noinspection PyUnboundLocalVariable,PyProtectedMember
def run(bm, variations, backwardPass=False, optimiser=ionbench.optimisers.scipy_optimisers.trustRegionReflective_scipy.run, filename=''):
    """
    Generate a profile likelihood style plot, reporting the fitted cost as a function of each fixed parameter.

    Parameters
    ----------
    bm : benchmarker
        A benchmarker function to explore.
    variations : list
        A list of parameter variations to apply for each parameter. It should be a list of length bm.n_parameters(), where the ith element is a list of variations to use to fix parameter i and fit the remaining parameters.
    backwardPass : bool, optional
        If False, the profile likelihood curve will be found by going left to right, using the optimised parameters from the left (lower variation) to initiate the optimisation. If True, it will travel right to left and the final name will be appended with 'B'. Both can be combined when plotting which can be advantageous to ensure a smooth profile likelihood plot. The default is False.
    optimiser: function, optional
        The optimiser to use for the profile likelihood plots. The default is trust region reflective.
    filename : string, optional
        A name to use to save the cost data. It will pickle the cost and variations and save them under [filename]_param[i].pickle. The default is '', in which case no data will be saved.

    Returns
    -------
    None.

    """
    bm.plotter = False
    if backwardPass:
        filename += 'B'  # New filename if solving backwards
    for i in range(bm.n_parameters()):
        # For each parameter...
        print('Working on parameter ' + str(i))
        costs = np.zeros(len(variations[i]))
        for j in range(len(variations[i])):
            # For each variation of that parameter...
            bm.reset()
            if backwardPass:
                varIndex = len(variations[i]) - j - 1
            else:
                varIndex = j
            var = variations[i][varIndex]
            # Create the profile manager, aligning on end points
            if j == 0:
                pm = ProfileManager(bm, i, bm._TRUE_PARAMETERS[i] * var, bm._TRUE_PARAMETERS)
            else:
                pm = ProfileManager(bm, i, bm._TRUE_PARAMETERS[i] * var, out)
            try:
                # Try to optimise the cost
                out = optimiser(pm)
                pm.MLE = bm._TRUE_PARAMETERS
                if pm.cost(out) >= 0.99 * pm.cost(
                        pm.sample()):  # If optimised cost not significantly better than unoptimised from default rates
                    pm.MLE = bm._TRUE_PARAMETERS
                    out = optimiser(pm)
            except Exception as e:  # pragma: no cover
                print(
                    'The optimisation caused an error (detailed below). In an attempt to recover, the profile likelihood will jump to the best guess.')
                print(e)
                out = pm.sample()

            costs[varIndex] = pm.cost(out)
            out = pm.set_params(out)
            print('Variation: ' + str(var))
            print('Cost found: ' + str(costs[varIndex]))
        if not filename == '':
            data = (variations[i], costs)
            with open(filename + '_param' + str(i) + '.pickle', 'wb') as f:
                pickle.dump(data, f)


# noinspection PyProtectedMember
def plot_profile_likelihood(modelType, filepath='', sharey=True, numPoints=51, debug=False):
    """
    Plot profile likelihood plots based on the pickled data in the current working directory.

    Parameters
    ----------
    modelType : string
        Type of model in benchmarker. This is used to load the benchmarker (options are hh, mm, ikr, ikur) and as the filename to load the pickled data [modelType]_param[i].pickle for i in range(numberToPlot)
    filepath : string, optional
        The filepath to a directory to save the figure inside. The default is '', in which case the figure will not be saved.
    sharey : bool, optional
        If True, all subplots will share the same y-axis. The default is True.
    numPoints: int, optional
        The number of points to use for the unoptimised cost slice. The default is 50.
    debug : bool, optional
        If True, extra plots will be drawn to separate the forwards and backwards passes of the profile likelihood optimisation. The default is False.
    """
    # Initialise a benchmarker
    if modelType == 'hh':  # pragma: no cover
        bm = ionbench.problems.staircase.HH()
        title = 'Staircase HH'
    elif modelType == 'mm':  # pragma: no cover
        bm = ionbench.problems.staircase.MM()
        title = 'Staircase MM'
    elif modelType == 'ikr':  # pragma: no cover
        bm = ionbench.problems.loewe2016.IKr()
        title = r'Loewe $\mathrm{I}_\mathrm{Kr}$'
    elif modelType == 'ikur':  # pragma: no cover
        bm = ionbench.problems.loewe2016.IKur()
        title = r'Loewe $\mathrm{I}_\mathrm{Kur}$'
    elif modelType == 'ina':  # pragma: no cover
        bm = ionbench.problems.moreno2016.INa()
        title = r'Moreno $\mathrm{I}_\mathrm{Na}$'
    else:  # pragma: no cover
        bm = None
    # Use scale factors so only variations needs to be specified as the parameters
    bm.useScaleFactors = True
    numberToPlot = bm.n_parameters()
    if sharey:
        rowSize = 5
    else:
        rowSize = 3
    perturbedCosts = []
    fig, axs = plt.subplots(int(np.ceil(numberToPlot/rowSize)), rowSize, figsize=(7.5, np.ceil(numberToPlot/rowSize)*1.5), layout='tight', sharey=sharey, sharex=True)
    minCost = np.inf
    maxCost = 0
    for i in range(numberToPlot):
        # Load the pickled data
        with open(modelType + '_param' + str(i) + '.pickle', 'rb') as f:
            variationsA, costsA = pickle.load(f)
            variations, costs = np.copy(variationsA), np.copy(costsA)
            # Central cost can be slightly off due to OS differences between data generation (Windows) and profile likelihood generation (Linux)
            if 'loewe' in bm.NAME or 'moreno' in bm.NAME:
                costs[variations == 1] = 0
        try:
            with open(modelType + 'B_param' + str(i) + '.pickle', 'rb') as f:
                variationsB, costsB = pickle.load(f)
            if len(variationsB) == len(variationsA):
                costs = np.array([min(costs[i], costsB[i]) for i in range(len(costs))])
        except FileNotFoundError as e:  # pragma: no cover
            print(e)
            pass

        if minCost > np.min(costs[costs > 0]):
            minCost = np.min(costs[costs > 0])
        if maxCost < np.max(costs[np.logical_and(costs < 1e5, costs > 0)]):
            maxCost = np.max(costs[np.logical_and(costs < 1e5, costs > 0)])

        # Calculate cost threshold
        lowCost = np.interp(0.95, variations, costs)
        if lowCost < 1e5:
            perturbedCosts.append(lowCost)
        highCost = np.interp(1.05, variations, costs)
        if highCost < 1e5:
            perturbedCosts.append(highCost)

        # Plot the profile likelihood
        axs[i//rowSize, i % rowSize].semilogy(variations, costs, label='Optimised' if i == 0 else None, zorder=1)

        # If debug, plot forwards and backwards cost separately
        if debug:
            axs[i//rowSize, i % rowSize].semilogy(variations, costsA, label='Forwards' if i == 0 else None, zorder=2, linestyle='dashed')
            try:
                if len(variationsB) == len(variationsA):
                    axs[i//rowSize, i % rowSize].semilogy(variations, costsB, label='Backwards' if i == 0 else None, zorder=3, linestyle='dotted')
            except NameError:  # pragma: no cover
                pass
        axs[i//rowSize, i % rowSize].set_xlabel(f'Parameter {i+1}')
        if i % rowSize == 0:
            axs[i//rowSize, i % rowSize].set_ylabel('Cost')

        if not sharey:
            # Penalty costs need to be removed now because sharey=False is a bit lazy with ticks
            ylim = (np.min(costs), np.max([costs[costs < 1e5]]))
            axs[i//rowSize, i % rowSize].set_ylim(ylim)
    # Get cost threshold
    perturbedCosts = np.array(perturbedCosts)
    threshold = np.min(perturbedCosts[np.logical_and(perturbedCosts > 1e-10, perturbedCosts > bm.cost(bm.input_parameter_space(bm._TRUE_PARAMETERS)))])
    thresholdIndex = np.argmin(np.abs(perturbedCosts - threshold))
    thresholdPlot = thresholdIndex//2
    if thresholdIndex % 2 == 0:
        x = 0.95
    else:
        x = 1.05
    # Round cost threshold up to 3 sig fig
    print(f'Threshold before: {threshold}')
    threshold = float(np.format_float_scientific(threshold + 0.5 * 10 ** (np.floor(np.log10(threshold)) - 2), 2))
    print(f'Threshold after: {threshold}')


    axs[thresholdPlot//rowSize, thresholdPlot % rowSize].axvline(x, color='black', linestyle='dotted', label='Threshold')

    # Generate lots of tick options
    ylim = [10**(np.log10(minCost)-axs[0, 0].margins()[1]*(np.log10(maxCost)-np.log10(minCost))), 10**(np.log10(maxCost)+axs[0, 0].margins()[1]*(np.log10(maxCost)-np.log10(minCost)))]
    if maxCost/minCost < 1e3:
        # Small variation, finer scale is needed
        loc = mpl.ticker.LogLocator(numticks=20, subs=range(1, 10))
        possibleTicks = loc.tick_values(*ylim)

        # Find the first value in possibleTicks that is greater than ylim[0] and the last value that is less than ylim[1]
        yticks = []
        for i in range(len(possibleTicks)):
            if possibleTicks[i] > ylim[0]:
                yticks.append(possibleTicks[i])
                break
        for i in range(len(possibleTicks)-1, -1, -1):
            if possibleTicks[i] < ylim[1]:
                yticks.append(possibleTicks[i])
                break
        yticks = np.array(yticks)

    for i in range(numberToPlot):
        # Plot unoptimised cost slice
        costs = np.zeros(numPoints)
        x = np.linspace(np.min(variations), np.max(variations), numPoints)
        for j in range(numPoints):
            p = bm.input_parameter_space(bm._TRUE_PARAMETERS)
            p[i] = x[j]
            costs[j] = bm.cost(p)
        axs[i//rowSize, i % rowSize].semilogy(x, costs, label='Unoptimised' if i == 0 else None, zorder=0, scaley=False)
        # Add horizontal line at cost threshold
        axs[i//rowSize, i % rowSize].axhline(threshold, color='black', linestyle='dotted')

    for i in range(numberToPlot, int(np.ceil(numberToPlot/rowSize))*rowSize):
        axs[i//rowSize, i % rowSize].axis('off')

    if sharey:
        if maxCost/minCost < 1e3:
            axs[0, 0].set_ylim(ylim)
            axs[0, 0].set_yticks(yticks)
        else:
            ylim = [10 ** np.floor(np.log10(ylim[0])), 10 ** np.ceil(np.log10(ylim[1]))]
            axs[0, 0].set_ylim(ylim)
            axs[0, 0].set_yticks(ylim)

    # Turn off minor ticks
    for ax in axs.flatten():
        ax.yaxis.set_minor_formatter(mpl.ticker.NullFormatter())
        ax.yaxis.set_minor_locator(mpl.ticker.NullLocator())
        if not sharey:
            # Set major ticks, don't bother with the log scale
            ax.yaxis.set_major_locator(mpl.ticker.MaxNLocator(nbins=4))

    # add labels
    fig.legend(loc='upper center', bbox_to_anchor=(0.5, 0))
    fig.suptitle(f'Profile likelihoods for {title}')
    # Save and show figure
    if filepath != '':
        fig.savefig(os.path.join(filepath, f'profileLikelihood-{modelType}{"-nonshared" if not sharey else ""}{"-debug" if debug else ""}'),
                    dpi=300, bbox_inches='tight')
    fig.show()

    # Print final cost threshold
    print(f'Threshold calculated as {threshold}')



import os
import ionbench


def test_multistart():
    # Check multistart runs (using lm scipy optimiser) and improves cost in all cases
    numStarts = 3
    bm = ionbench.problems.staircase.HH()
    bm.plotter = False
    initParams = bm.sample(n=numStarts)
    costs = []
    for x0 in initParams:
        costs.append(bm.cost(x0))
    outs = ionbench.utils.multistart(ionbench.optimisers.scipy_optimisers.nelderMead_scipy.run, bm, initParams, 'test', maxIter=5)
    for i in range(numStarts):
        assert bm.cost(outs[i]) <= costs[i]
        os.remove('test_run' + str(i) + '.pickle')




import ionbench
import scipy
import numpy as np


def least_squares(bm, x0, debug, method, maxIter, **kwargs):
    """
    Run a scipy least_squares optimiser. Will use cached signed_error and jacobians and return the scipy optimisation output.
    Parameters
    ----------
    bm : Benchmarker
        A benchmarker object.
    x0 : list
        Initial parameter vector from which to start optimisation. If None, a randomly sampled parameter vector is retrieved from bm.sample().
    debug : bool
        If True, prints out the cost and parameters found by the algorithm.
    method : str
        The method to pass into least_squares.
    maxIter : int
        Maximum number of iterations.
    kwargs : dict, optional
        Additional keyword arguments to pass to least_squares.

    Returns
    -------
    out : scipy.optimize.OptimizeResult
        The result of the optimisation.
    """
    def grad(p):
        return bm.grad(p, residuals=True)

    if x0 is None:
        x0 = bm.sample()
        if debug:
            print('Sampling x0')
            print(x0)

    if debug:
        verbose = 2
    else:
        verbose = 1
    if bm.parametersBounded and method != 'lm':
        bounds = minimize_bounds(bm)
    else:
        bounds = (-np.inf, np.inf)
    out = scipy.optimize.least_squares(bm.signed_error, x0, method=method, bounds=bounds, jac=grad, verbose=verbose, max_nfev=maxIter,
                                       **kwargs)

    if out.nfev >= maxIter:
        bm.set_max_iter_flag()

    if debug:
        print(f'Cost of {out.cost} found at:')
        print(out.x)

    return out


def minimize_bounds(bm):
    """
    Map the benchmarker bounds to the scipy bounds format.
    Parameters
    ----------
    bm : Benchmarker
        A benchmarker object.
    Returns
    -------
    bounds : scipy.optimize.Bounds
        Scipy bounds object which restricts scipy optimisation to remain inside parameter bounds.
    """
    lb = bm.input_parameter_space(bm.lb)
    ub = bm.input_parameter_space(bm.ub)
    bounds = scipy.optimize.Bounds(lb=lb, ub=ub, keep_feasible=True)
    return bounds



import ionbench.problems.staircase
import scipy.optimize
from ionbench.utils.scipy_setup import minimize_bounds


# noinspection PyShadowingNames
def run(bm, x0=None, xtol=1e-4, ftol=1e-4, maxIter=1000, maxfev=20000, debug=False):
    """
    Runs Nelder-Mead optimiser from Scipy. Bounds are automatically loaded from the benchmarker if present.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is None, in which case a randomly sampled parameter vector is retrieved from bm.sample().
    xtol : float, optional
        Tolerance in parameters. Used as a termination criterion. The default is 1e-4.
    ftol : float, optional
        Tolerance in cost. Used as a termination criterion. The default is 1e-4.
    maxIter : int, optional
        Maximum number of iterations of Nelder-Mead to use. The default is 1000.
    maxfev : int, optional
        Maximum number of cost function evaluations. The default is 20000.
    debug : bool, optional
        If True, prints out the cost and parameters found by the algorithm. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified by Nelder-Mead.
    """
    if x0 is None:
        x0 = bm.sample()
        if debug:
            print('Sampling x0')
            print(x0)

    if bm.parametersBounded:
        bounds = minimize_bounds(bm)
        out = scipy.optimize.minimize(bm.cost, x0, method='nelder-mead', options={'disp': debug, 'xatol': xtol, 'fatol': ftol, 'maxiter': maxIter, 'maxfev': maxfev}, bounds=bounds)
    else:
        out = scipy.optimize.minimize(bm.cost, x0, method='nelder-mead', options={'disp': debug, 'xatol': xtol, 'fatol': ftol, 'maxiter': maxIter, 'maxfev': maxfev})

    if out.nit >= maxIter or out.nfev >= maxfev:
        bm.set_max_iter_flag()

    if debug:
        print(f'Cost of {out.fun} found at:')
        print(out.x)

    bm.evaluate()
    return out.x


# noinspection PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Balser1990
    modNum = 2 -> Davies2012
    modNum = 4 -> Moreno2016

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Balser1990.

    """

    if modNum == 1:
        mod = ionbench.modification.Balser1990()
    elif modNum == 2:
        mod = ionbench.modification.Davies2012()
    elif modNum == 3:
        mod = ionbench.modification.Moreno2016()
    else:
        mod = ionbench.modification.Empty(name='nelderMead_scipy')
    return mod



import pints
import ionbench
from ionbench.utils import classes_pints


# noinspection PyShadowingNames
def run(bm, x0=None, maxIter=1000, debug=False):
    """
    Runs PSO (Particle Swarm Optimisation) from Pints using a benchmarker.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is None, in which case a randomly sampled parameter vector is retrieved from bm.sample().
    maxIter : int, optional
        Number of iterations of PSO to run. The default is 1000.
    debug : bool, optional
        If True, logging messages are printed every iteration. Otherwise, the default of every iteration for the first 3 and then every 20 iterations. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified by PSO.

    """

    model, opt = classes_pints.pints_setup(bm, x0, pints.PSO, maxIter, debug)
    x, f = opt.run()
    if opt.iterations() >= maxIter:
        model.bm.set_max_iter_flag()
    model.bm.evaluate()
    return x


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    No modification for this optimiser. Will use an empty modification.

    Returns
    -------
    mod : modification
        Empty modification

    """
    mod = ionbench.modification.Empty(name='pso_pints')
    return mod



"""
This module contains the PSO optimiser from Cabo 2022.
Little information on the algorithm is given in the paper but the code that was used is provided at https://github.com/kkentzo/pso
The code provides many options for the algorithm, which are not described in the paper. We use the defaults, also stated here.
There is a choice of topologies for determining the global acceleration term, in which case we use the ring topology (PSO_NHOOD_RING)
There is a choice of inertia weights, either linearly decreasing (user specified limits) or constant (0.7298). We use linearly decreasing from 0.7298 to 0.3 over maxIter iterations.
We use the internally calculated population size formula.
We assume the coefficients c1 and c2 are maintained at their defaults (1.496).
We assume the particle positions are clamped to the bounds.
The initial position sampling described would ignore the point x0 so cannot be used in ionBench. We use the ionBench default instead.
The initial velocity sampling can be used.
"""
import numpy as np
import ionbench
import copy


# noinspection PyShadowingNames
def run(bm, x0=None, maxIter=1000, debug=False):
    """
    Runs the PSO algorithm defined by Cabo 2022.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. Population is generated by randomly perturbing this initial guess +-50%, then applying the appropriate bounds. If x0=None (the default), then the population will be sampled uniformly between the bounds.
    maxIter : int, optional
        Maximum number of iterations. The default is 1000.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    if x0 is None:
        x0 = bm.sample()

    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    # noinspection PyShadowingNames
    class Particle(ionbench.utils.particle_optimisers.Particle):
        def __init__(self):
            self.x0 = x0
            super().__init__(bm, cost_func, x0)

        def set_velocity(self):
            """
            The velocity algorithm used by Cabo 2022 takes two sampled positions. Sets the particle position to one of them, set the velocity to point away from the other.
            """
            copied_position = np.copy(self.position)
            self.set_position(self.x0)
            self.velocity = (self.position - copied_position) / 2
            # Reset position
            self.position = copied_position

        def clamp(self):
            """
            Clamp parameters to the input parameter space. Also sets the velocity to 0 in any clamped parameters.
            """
            for i in range(len(self.position)):
                if self.position[i] < 0:  # pragma: no cover
                    self.position[i] = 0
                    self.velocity[i] = 0
                elif self.position[i] > 1:  # pragma: no cover
                    self.position[i] = 1
                    self.velocity[i] = 0

    # noinspection PyShadowingNames
    def best_in_ring(particleList):
        """
        Finds the best cost and position in the ring topology for each particle. Returns the positions of the best neighbouring particle that each particle can see.

        Parameters
        ----------
        particleList : list
            List of particles.

        Returns
        -------
        ringPos : list
            List of positions of the best neighbouring particle that each particle can see.
        """
        costs = [p.bestCost for p in particleList]
        ringPos = [p.bestPosition for p in particleList]
        ringCosts = copy.copy(costs)
        for i in range(len(particleList)):
            costLow = costs[(i - 1) % len(particleList)]
            costHigh = costs[(i + 1) % len(particleList)]
            if costLow < ringCosts[i]:
                ringCosts[i] = costLow
                ringPos[i] = particleList[(i - 1) % len(particleList)].bestPosition
            if costHigh < ringCosts[i]:
                ringCosts[i] = costHigh
                ringPos[i] = particleList[(i + 1) % len(particleList)].bestPosition
        return ringPos

    # Set population size n
    n = int(10 + 2 * np.sqrt(bm.n_parameters()))
    # Initial population
    particleList = [Particle() for _ in range(n)]

    Gcost = [np.inf] * maxIter  # Best cost ever
    Gpos = [None] * maxIter  # Position of best cost ever
    L = None
    for L in range(maxIter):
        if L > 0:
            Gcost[L] = Gcost[L - 1]
            Gpos[L] = Gpos[L - 1]

        if debug:
            print('-------------')
            print(f'Beginning population: {L}')
            print(f'Best cost so far: {Gcost[L]}')
            print(f'Found at position: {Gpos[L]}')

        for p in particleList:
            p.set_cost()
            if p.currentCost < Gcost[L]:
                Gcost[L] = p.currentCost
                Gpos[L] = np.copy(p.position)

        # Renew velocities
        c1 = 1.496
        c2 = 1.496
        if L < 1000:  # Dependence of w on maxIter has been removed
            w = 0.7298 - (0.7298-0.3)*L/1000
        else:
            w = 0.3
        ringPos = best_in_ring(particleList)
        for i, p in enumerate(particleList):
            localAcc = c1 * np.random.rand() * (p.bestPosition - p.position)
            globalAcc = c2 * np.random.rand() * (ringPos[i] - p.position)
            p.velocity = w * p.velocity + localAcc + globalAcc
        if debug:
            print("Velocities renewed")
        # Move positions
        for p in particleList:
            p.position += p.velocity
            p.clamp()

        if debug:
            print("Positions renewed")
            print(f'Finished population {L}')
            print(f'Best cost so far: {Gcost[L]}')
            print(f'Found at position: {Gpos[L]}')

        if bm.is_converged():
            break

    if L >= maxIter-1:
        bm.set_max_iter_flag()

    bm.evaluate()
    return Particle().untransform(Gpos[L])


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Cabo2022

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Cabo2022.

    """
    mod = ionbench.modification.Cabo2022()
    return mod



"""
This module describes the pattern search algorithm from Kohjitani et al. 2022.
This algorithm is a simple pattern search algorithm that explores neighbouring points around a base point, and moves to the best neighbouring point if an improvement is found.
The step size is reduced if no improvements are found in the neighbouring points.
The algorithm terminates when the step size is sufficiently small.
"""
import ionbench
import numpy as np


# noinspection PyShadowingNames
def run(bm, x0=None, maxIter=1000, CrtStp=2e-5, Stp=1 / 100, RedFct=1 / 4, maxfev=20000, debug=False):
    """
    Runs the pattern search algorithm from Kohjitani et al. 2022.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is None, in which case a randomly sampled parameter vector is retrieved from bm.sample().
    maxIter : int, optional
        Maximum number of iterations before termination.
    CrtStp : float, optional
        The minimum step size. If there are no improvements within CrtStp, the optimisation terminates. The default is 2e-5.
    Stp : float, optional
        Initial step size. The default is 1/100.
    RedFct : float, optional
        The reduction factor. If the center point is better than its neighbours, the step size is scaled by this reduction factor. The default is 1/4.
    maxfev : int, optional
        Maximum number of cost function evaluations. The default is 20000.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    funcCounter = 0
    iterCounter = 0

    # noinspection PyShadowingNames
    def explore(BP, Stp):
        """
        Explores neighbouring points to the base point BP.
        Parameters
        ----------
        BP : numpy array
            The base point around which to explore. Check cost by increasing and decreasing each parameter by Stp.
        Stp : float
            The step size.

        Returns
        -------
        foundImprovement : bool
            True if an improvement is found, False otherwise.
        NP : numpy array
            The new base point, if an improvement is found. Otherwise, the same as BP.
        """
        foundImprovement = False
        NP = np.copy(BP)
        MSE = cost_func(tuple(BP))  # No real computation cost from this thanks to the cache
        for i in range(bm.n_parameters()):
            home = BP[i]
            NP[i] = home + Stp  # Positive Step
            MSEp = cost_func(tuple(NP))  # Positive MSE
            NP[i] = home - Stp  # Negative Step
            MSEn = cost_func(tuple(NP))  # Negative MSE
            minMSE = min(MSEp, MSEn)  # MSE in the best direction (positive or negative Stp)
            if minMSE < MSE:  # If improvement found
                if MSEp < MSEn:  # If positive step is better
                    NP[i] = home + Stp  # Take positive step
                else:  # If negative step is better
                    NP[i] = home - Stp  # Take negative step
                MSE = minMSE  # Either way, record new MSE
                foundImprovement = True
            else:  # If no improvement
                NP[i] = home  # Restore center point
        return foundImprovement, NP

    if x0 is None:
        x0 = bm.sample()

    BP = x0  # Set initial base point
    NP = np.copy(BP)
    funcCounter += 1
    while Stp > CrtStp:  # Stop when step size is sufficiently small
        # Explore neighbouring points
        if debug:
            print("------------")
            print(f'Current step size: {Stp}')
            print(f'Cost: {cost_func(tuple(BP))}')
        improvementFound, NP = explore(BP, Stp)  # Explore neighbouring points
        funcCounter += 4
        while improvementFound:
            iterCounter += 1
            if debug:
                print(f'Improvement Found? {improvementFound}')
            BP = np.copy(NP)  # Move to new improved point
            improvementFound, NP = explore(BP, Stp)  # Explore neighbouring points
            funcCounter += 4
            if funcCounter > maxfev:
                print("Exceeded maximum number of function evaluations.")
                bm.set_max_iter_flag()
                bm.evaluate()
                return NP
            if iterCounter > maxIter:
                print("Exceeded maximum number of iterations.")
                bm.set_max_iter_flag()
                bm.evaluate()
                return NP

            if bm.is_converged():
                bm.evaluate()
                return NP

        Stp *= RedFct  # Decrease step size if all neighbouring points are worse

    bm.evaluate()
    return NP


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Kohjitani2022

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Kohjitani2022.

    """
    mod = ionbench.modification.Kohjitani2022()
    return mod



import os

import ionbench
import numpy as np
import matplotlib.pyplot as plt


def test_profile_likelihood(monkeypatch):
    # Generate profile likelihood plots (don't actually plot). Only checks for crashes/error
    def var(x):
        return np.linspace(1 - x, 1 + x, 3)

    def savefig(*args, **kwargs):
        pass

    monkeypatch.setattr(plt, 'show', lambda: None)
    monkeypatch.setattr(plt, 'savefig', savefig)
    variations = [var(0.2)] * 9
    bm = ionbench.problems.staircase.HH()
    ionbench.uncertainty.profile_likelihood.run(bm, variations, filename='hh')
    ionbench.uncertainty.profile_likelihood.run(bm, variations, filename='hh', backwardPass=True)
    ionbench.uncertainty.profile_likelihood.plot_profile_likelihood(modelType='hh', debug=True)
    ionbench.uncertainty.profile_likelihood.plot_profile_likelihood(modelType='hh', sharey=False, debug=True)
    for i in range(9):
        os.remove(f'hh_param{i}.pickle')
        os.remove(f'hhB_param{i}.pickle')



"""
The module provides the stochastic search algorithm from Vanier et al. 1999.
This algorithm was previously described in Foster et al. 1993.
We use the hyperparameters given by Vanier et al. 1999.
Note that while Vanier et al. 1999 describes the variance as decreasing linearly, we use a geometric decrease as described in Foster et al. 1993.
"""
import numpy as np
import ionbench


# noinspection PyShadowingNames
def run(bm, x0=None, varInit=0.5, varMin=0.05, varCont=0.95, maxIter=1000, debug=False):
    """
    Stochastic search from Vanier et al. 1999.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. If x0=None (the default), then the population will be sampled from the benchmarker problems .sample() method.
    varInit : float, optional
        The initial variance. The default is 0.5.
    varMin : float, optional
        The minimum variance. Once the variance is decreased below this minimum, it is reset to its initial value of varInit. The default is 0.05.
    varCont : float, optional
        The variance contraction rate. Every iteration the current variance is varCont times the previous. The default is 0.95.
    maxIter : int, optional
        Maximum number of iterations. The default is 1000.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters found.

    """
    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    if x0 is None:
        # sample initial point
        x0 = bm.sample()
    var = varInit
    x0_cost = cost_func(x0)
    if debug:
        print(f'Starting cost is {x0_cost}')
    i = None
    for i in range(maxIter):
        trial = x0 + np.random.normal(loc=np.zeros(bm.n_parameters()), scale=np.sqrt(var))*(bm.input_parameter_space(bm.ub)-bm.input_parameter_space(bm.lb))
        trial = bm.clamp_parameters(trial)
        trial_cost = cost_func(trial)
        if debug:
            print(x0)
            print(trial)
            print(f'var: {var}, trial cost: {trial_cost}')
        if trial_cost < x0_cost:
            x0 = trial
            x0_cost = trial_cost
            if debug:
                print('Found improvement')
        var *= varCont
        if var < varMin:
            var = varInit
        if bm.is_converged():
            break

    if i >= maxIter-1:
        bm.set_max_iter_flag()

    if debug:
        print('Complete')
        print(f'Final cost is {x0_cost}')

    bm.evaluate()
    return x0


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Vanier1999

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Vanier1999.

    """
    mod = ionbench.modification.Vanier1999()
    return mod



"""
Contains the multistart function (referenced as ionbench.multistart) which can perform multiple runs of the same optimisation function starting at different initial conditions.
"""
import ionbench.utils.cache


def multistart(opt, bm, initParams, filename, **kwargs):
    """
    Run an optimiser multiple times from different starting location.

    Parameters
    ----------
    opt : function
        A function which takes inputs of a benchmarker bm, a vector of initial parameters x0, and returns a vector of optimised parameters. All ionBench optimisers have a .run() function which satisfies these requirements.
    bm : Benchmarker
        A benchmarker problem to use for the optimisation.
    initParams : list
        A list of initial parameter vectors. The first x0 parameters used will be initParams[0].
    filename : string
        A filename to identify this optimisation run. The benchmarker tracker objects will store their data in '[filename]_run[i].pickle' for i in range(len(initParams)).
    kwargs
        Any additional keyword arguments to pass into the opt() call.

    Returns
    -------
    outs : list
        A list of optimised parameters. The same size as initParams.
    """
    outs = []
    for i in range(len(initParams)):
        out = opt(bm, x0=initParams[i], **kwargs)
        if not filename == '':
            bm.tracker.save(filename + '_run' + str(i) + '.pickle')
        print(out)
        outs.append(out)
        bm.reset(fullReset=False)
        ionbench.utils.cache.clear_all_caches()
    return outs



"""
Contains the two benchmarker problems from Loewe et al. 2016, IKr and IKur. These both inherit from LoeweBenchmarker which itself inherits from benchmarker.Benchmarker.
generate_data(modelType) will generate the data for either the IKr or IKur problem and store it in the data directory.
"""
import ionbench
import myokit
import myokit.lib.hh
import os
import numpy as np
import csv


class LoeweBenchmarker(ionbench.benchmarker.Benchmarker):
    def __init__(self, states):
        # Benchmarker
        p = self.protocol()
        self.TIMESTEP = 0.5  # Timestep in data between points
        self.T_MAX = p.characteristic_time()  # Length of protocol
        # Parameter bounds (hidden from user)
        self._LOWER_BOUND = np.array([self._TRUE_PARAMETERS[i] * 0.1 if self.STANDARD_LOG_TRANSFORM[i] else
                                      self._TRUE_PARAMETERS[i] - 60 for i in range(self.n_parameters())])
        self._UPPER_BOUND = np.array([self._TRUE_PARAMETERS[i] * 10 if self.STANDARD_LOG_TRANSFORM[i] else
                                      self._TRUE_PARAMETERS[i] + 60 for i in range(self.n_parameters())])
        # Parameter bounds (available to user)
        self.lb = np.copy(self._LOWER_BOUND)
        self.ub = np.copy(self._UPPER_BOUND)
        # Rate bounds
        self.RATE_MIN = 1.67e-5
        self.RATE_MAX = 1e3

        # Myokit
        parameters = [self._PARAMETER_CONTAINER + '.p' + str(i + 1) for i in range(self.n_parameters())]
        self._ANALYTICAL_MODEL = myokit.lib.hh.HHModel(model=self._MODEL, states=states, parameters=parameters,
                                                       current=self._OUTPUT_NAME, vm='membrane.V')
        self.sim = myokit.lib.hh.AnalyticalSimulation(self._ANALYTICAL_MODEL, protocol=p)
        if self.sensitivityCalc:
            # ODE solver
            sens = ([self._OUTPUT_NAME], parameters)
            self.simSens = myokit.Simulation(self._MODEL, sensitivities=sens, protocol=p)
            self.simSens.set_tolerance(*self._TOLERANCES)
        else:
            self.simSens = None
        super().__init__()

    def sample(self, n=1):
        """
        Sample parameters for the Loewe 2016 problems.

        Parameters
        ----------
        n : int, optional
            Number of parameter vectors to sample. The default is 1.

        Returns
        -------
        params : list
            If n=1, then params is the vector of parameters. Otherwise, params is a list containing n parameter vectors.

        """
        params = [None] * n
        for i in range(n):
            param = [None] * self.n_parameters()
            for j in range(self.n_parameters()):
                if self.STANDARD_LOG_TRANSFORM[j]:
                    param[j] = self._TRUE_PARAMETERS[j] * 10 ** np.random.uniform(-1, 1)
                else:
                    param[j] = self._TRUE_PARAMETERS[j] + np.random.uniform(-60, 60)
            params[i] = self.input_parameter_space(param)  # Generates a copy
        if n == 1:
            return params[0]
        else:
            return params

    @staticmethod
    def protocol():
        """
        Add the protocol from Loewe et al. 2016. This protocol consists of 20ms at -80mV, a variable height step between 50mV and -70mV (in steps of 10mV) for 400ms and then 400ms at -110mV before repeating with a new height for the central step (in order of decreasing voltage).

        Returns
        -------
        p : myokit.Protocol
            A myokit.Protocol for the voltage clamp protocol from Loewe et al. 2016.

        """
        p = myokit.Protocol()
        vsteps = []
        for i in range(13):
            vsteps.append(-80)
            vsteps.append(50 - i * 10)
            vsteps.append(-110)
        durations = [20, 400, 400] * 13
        for i in range(len(vsteps)):
            p.add_step(vsteps[i], durations[i])
        return p


class IKr(LoeweBenchmarker):
    """
    The Loewe 2016 IKr benchmarker.

    The benchmarker uses the Courtemanche 1998 IKr model with a simple step protocol.

    Its parameters are specified as reported in Loewe et al. 2016 with the true parameters being the same as the default and the center of the sampling distribution.
    """

    def __init__(self, sensitivities=False):
        print('Initialising Loewe 2016 IKr benchmark')
        # Benchmarker
        self.NAME = "loewe2016.ikr"
        self._TRUE_PARAMETERS = np.array([3e-4, 14.1, 5, 3.3328, 5.1237, 1, 14.1, 6.5, 15, 22.4, 0.029411765, 138.994])
        self.STANDARD_LOG_TRANSFORM = (True, False, True, False, True, True, False, True, False, True, True, True)
        self._RATE_FUNCTIONS = (lambda p, V: p[0] * (V + p[1]) / (1 - np.exp((V + p[1]) / (-p[2]))),
                                lambda p, V: 7.3898e-5 * (V + p[3]) / (np.exp((V + p[3]) / p[4]) - 1))  # Used for rate bounds
        self.sensitivityCalc = sensitivities
        self.COST_THRESHOLD = 2.95e-6

        # Myokit
        self._MODEL = myokit.load_model(os.path.join(ionbench.DATA_DIR, 'loewe2016', 'courtemanche-1998-ikr.mmt'))
        self._OUTPUT_NAME = 'ikr.IKr'
        self._PARAMETER_CONTAINER = 'ikr'
        self._TOLERANCES = (1e-6, 1e-6)
        self.load_data(dataPath=os.path.join(ionbench.DATA_DIR, 'loewe2016', 'ikr.csv'))
        super().__init__(states=['ikr.xr'])
        print('Benchmarker initialised')


class IKur(LoeweBenchmarker):
    """
    The Loewe 2016 IKur benchmarker.

    The benchmarker uses the Courtemanche 1998 IKur model with a simple step protocol.

    Its parameters are specified as reported in Loewe et al. 2016 with the true parameters being the same as the default and the center of the sampling distribution.
    """

    def __init__(self, sensitivities=False):
        print('Initialising Loewe 2016 IKur benchmark')
        # Benchmarker
        self.NAME = "loewe2016.ikur"
        self._TRUE_PARAMETERS = np.array(
            [0.65, 10, 8.5, 30, 59, 2.5, 82, 17, 30.3, 9.6, 3, 1, 21, 185, 28, 158, 16, 99.45, 27.48, 3, 0.005, 0.05,
             15, 13, 138.994])
        self.STANDARD_LOG_TRANSFORM = (True, False, True, False, True, False, False, True, False, True, True, True,
                                       False, False, True, False, True, False, True, True, False, True, False, True,
                                       True)
        self._RATE_FUNCTIONS = (
            lambda p, V: p[0] / (np.exp((V + p[1]) / -p[2]) + np.exp((V - p[3]) / -p[4])),
            lambda p, V: 0.65 / (p[5] + np.exp((V + p[6]) / p[7])),
            lambda p, V: p[11] / (p[12] + np.exp((V - p[13]) / -p[14])),
            lambda p, V: np.exp((V - p[15]) / p[16]))  # Used for rate bounds
        self.sensitivityCalc = sensitivities
        self.COST_THRESHOLD = 2.80e-7

        # Myokit
        self._MODEL = myokit.load_model(os.path.join(ionbench.DATA_DIR, 'loewe2016', 'courtemanche-1998-ikur.mmt'))
        self._OUTPUT_NAME = 'ikur.IKur'
        self._PARAMETER_CONTAINER = 'ikur'
        self._TOLERANCES = (1e-8, 1e-8)
        self.load_data(dataPath=os.path.join(ionbench.DATA_DIR, 'loewe2016', 'ikur.csv'))
        super().__init__(states=['ikur.ua', 'ikur.ui'])
        print('Benchmarker initialised')


# noinspection PyProtectedMember
def generate_data(modelType):  # pragma: no cover
    """
    Generate the data files for the Loewe 2016 benchmarker problems. The true parameters are the same as the default for these benchmark problems.

    Parameters
    ----------
    modelType : string
        'ikr' to generate the data for the IKr benchmark problem. 'ikur' to generate the data for the IKur benchmark problem.

    Returns
    -------
    None.

    """
    modelType = modelType.lower()
    if modelType == 'ikr':
        bm = IKr()
    else:
        bm = IKur()
    bm.set_params(bm._TRUE_PARAMETERS)
    bm.set_steady_state(bm._TRUE_PARAMETERS)
    out = bm.solve_model(np.arange(0, bm.T_MAX, bm.TIMESTEP), continueOnError=False)
    with open(os.path.join(ionbench.DATA_DIR, 'loewe2016', modelType + '.csv'), 'w', newline='') as csvfile:
        writer = csv.writer(csvfile, delimiter=',')
        writer.writerows(map(lambda x: [x], out))



import os
from ionbench import benchmarker
from ionbench import problems
from ionbench import optimisers
from ionbench import modification
from ionbench import uncertainty
from ionbench import utils

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))

DATA_DIR = os.path.join(ROOT_DIR, "data")

OPT_SCIPY = ['ionbench.optimisers.scipy_optimisers.' + st + '_scipy' for st in ['conjugateGD', 'lm', 'nelderMead', 'powell', 'slsqp', 'trustRegionReflective']]

OPT_PINTS = ['ionbench.optimisers.pints_optimisers.' + st + '_pints' for st in ['cmaes', 'nelderMead', 'pso', 'snes', 'xnes', 'rprop']]

OPT_EXT = ['ionbench.optimisers.external_optimisers.' + st for st in ['curvilinearGD_Dokos2004', 'DE_Zhou2009', 'GA_Bot2012', 'GA_Cairns2017', 'GA_Gurkiewicz2007a', 'GA_Gurkiewicz2007b', 'GA_Smirnov2020', 'hybridPSOTRR_Loewe2016', 'hybridPSOTRRTRR_Loewe2016', 'NMPSO_Clausen2020', 'NMPSO_Liu2011', 'patternSearch_Kohjitani2022', 'ppso_Chen2012', 'pso_Cabo2022', 'PSO_Loewe2016', 'pso_Seemann2009', 'PSOTRR_Loewe2016', 'randomSearch_Vanier1999', 'SA_Vanier1999', 'SPSA_Spall1998', 'stochasticSearch_Vanier1999']]

OPT_ALL = OPT_SCIPY + OPT_PINTS + OPT_EXT

N_MOD_SCIPY = [2, 2, 3, 3, 1, 3]
N_MOD_PINTS = [3, 0, 0, 0, 0, 0]
N_MOD_EXT = [3, 1, 2, 1, 2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
N_MOD_ALL = N_MOD_SCIPY + N_MOD_PINTS + N_MOD_EXT

APP_SCIPY = []
for i in range(len(OPT_SCIPY)):
    for j in range(N_MOD_SCIPY[i]):
        APP_SCIPY.append({'module': OPT_SCIPY[i], 'modNum': j + 1, 'kwargs': {}})

APP_PINTS = []
for i in range(len(OPT_PINTS)):
    for j in range(N_MOD_PINTS[i]):
        APP_PINTS.append({'module': OPT_PINTS[i], 'modNum': j + 1, 'kwargs': {}})

APP_EXT = []
for i in range(len(OPT_EXT)):
    for j in range(N_MOD_EXT[i]):
        APP_EXT.append({'module': OPT_EXT[i], 'modNum': j + 1, 'kwargs': {}})

APP_ALL = APP_SCIPY + APP_PINTS + APP_EXT
APP_UNIQUE = [APP_ALL[i] for i in range(len(APP_ALL)) if i not in [3, 8, 9, 12, 13, 16, 18, 19]]

cache_enabled = True



import pints
import ionbench
from ionbench.utils import classes_pints


# Limited information on the implementation given in Clausen 2020.


# noinspection PyShadowingNames
def run(bm, x0=None, maxIter=1000, debug=False):
    """
    Runs PSO (Particle Swarm Optimisation) from Pints and then Nelder Mead from Pints.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter vector from which to start optimisation. Default is None, in which case a randomly sampled parameter vector is retrieved from bm.sample().
    maxIter : int, optional
        Number of iterations of Nelder Mead to run after PSO. PSO is hardcoded to use a maximum of 1000 iterations. The maxIter flag is set using the maximum number of iterations of Nelder Mead. The default is 1000.
    debug : bool, optional
        If True, debug information will be printed, reporting more optimisation information. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified by PSO.

    """

    model, opt = classes_pints.pints_setup(bm, x0, pints.PSO, maxIter, debug)
    opt.set_max_iterations(1000)
    if debug:
        print('Beginning PSO')
    # Run the optimisation
    x, f = opt.run()
    if debug:
        print(f'PSO complete with best cost of {f}')
    # Create an optimisation controller
    model, opt = classes_pints.pints_setup(bm, x, pints.NelderMead, maxIter, debug, forceUnbounded=True)
    # Run the optimisation
    if debug:
        print('Beginning NM')
    x, f = opt.run()
    if opt.iterations() >= maxIter:
        model.bm.set_max_iter_flag()
    if debug:
        print(f'Nelder Mead complete with best cost of {f}')
    model.bm.evaluate()
    return x


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Clausen2020

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Clausen2020.

    """
    mod = ionbench.modification.Clausen2020()
    return mod



import numpy as np
import ionbench
import ionbench.utils.population_optimisers as pop_opt
import copy
import warnings


# noinspection PyShadowingNames
def run(bm, x0=None, nGens=50, popSize=50, crossoverProb=0.5, debug=False):
    """
    Runs the genetic algorithm from Gurkiewicz et al. 2007. This uses tournament selection (tournament size of 2) and single point crossover. This version of the algorithm, labelled version a, draws a new sample for the mutated parameters, ignoring the previous value.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. Population is generated by randomly perturbing this initial guess +-50%, clamped to bounds if necessary. If x0=None (the default), then the population will be sampled using bm.sample().
    nGens : int, optional
        The number of generations to run the optimisation algorithm for. The default is 1000.
    popSize : int, optional
        The size of the population in each generation. If the popSize is less than twenty times the number of parameters, it will be increased to twenty times the number of parameters. The default is 0.
    crossoverProb : float, optional
        The probability of a crossover. Gurkiewicz 2007 uses 0.5 (the default) while Ben-Shalom 2012 uses 0.1.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """

    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    if popSize < 20 * bm.n_parameters():
        warnings.warn(
            'Too small a value specified for popSize. Gurkiewicz recommends using at least 20 times the number of parameters. popSize will be increased to this value.')
        popSize = 20 * bm.n_parameters()

    pop = pop_opt.get_pop(bm, x0, popSize, cost_func)

    gen = None
    for gen in range(nGens):
        elites = pop_opt.get_elites(pop, 1)
        if debug:
            print("------------")
            print(f'Gen {gen}, Best cost: {elites[0].cost}')

        # Tournament selection
        pop = pop_opt.tournament_selection(pop)

        # One Point Crossover
        pop = pop_opt.one_point_crossover(pop, bm, cost_func, crossoverProb=crossoverProb)

        # Mutation
        for i in range(popSize):
            for j in range(bm.n_parameters()):
                if np.random.rand() < 0.01:
                    pop[i].x[j] = bm.sample()[j]

        if debug:
            print(f'Finishing gen {gen}')

        # Find costs
        pop = pop_opt.find_pop_costs(pop)

        # Elitism
        pop = pop_opt.set_elites(pop, elites)

        if bm.is_converged():
            break

    if gen >= nGens - 1:
        bm.set_max_iter_flag()

    elites = pop_opt.get_elites(pop, 1)
    bm.evaluate()
    return elites[0].x


# noinspection PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Gurkiewicz2007
    modNum = 2 -> BenShalom2012

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Gurkiewicz2007.

    """

    if modNum == 1:
        mod = ionbench.modification.Gurkiewicz2007()
    elif modNum == 2:
        mod = ionbench.modification.BenShalom2012()
    else:
        mod = ionbench.modification.Empty(name='GA_Gurkiewicz2007')
    return mod



from functools import lru_cache
import copy
import ionbench

cachedFunctions = []


def cache():
    def decorator(func):
        func = lru_cache(maxsize=None)(func)
        cachedFunctions.append(func)
        return func
    return decorator


def clear_all_caches():
    for func in cachedFunctions:
        func.cache_clear()


def get_cached_cost(bm):
    """
    Returns a cached version of the bm.cost() function for the inputted benchmark.

    Parameters
    ----------
    bm : benchmarker
        Benchmarker object.

    Returns
    -------
    cost_func : function
        Cached cost function.
    """
    if ionbench.cache_enabled:
        @cache()
        def cached_func(p):
            return bm.cost(p)
    else:  # pragma: no cover
        def cached_func(p):
            return bm.cost(p)

    def cost_func(p):
        return cached_func(tuple(p))

    return cost_func


def get_cached_signed_error(bm):
    """
    Returns a cached version of the bm.signed_error() function for the inputted benchmark.

    Parameters
    ----------
    bm : benchmarker
        Benchmarker object.

    Returns
    -------
    signed_error : function
        Cached signed error (residuals) function.
    """

    if ionbench.cache_enabled:
        @cache()
        def cached_func(p):
            return bm.signed_error(p)
    else:  # pragma: no cover
        def cached_func(p):
            return bm.signed_error(p)

    def signed_error(p):
        return cached_func(tuple(p))

    return signed_error


def get_cached_grad(bm, **kwargs):
    """
    Returns a cached version of the bm.grad() function for the inputted benchmark.

    Parameters
    ----------
    bm : benchmarker
        Benchmarker object.
    kwargs : dict
        Keyword arguments to pass to the gradient function.

    Returns
    -------
    grad : function
        Cached grad function.
    """

    if ionbench.cache_enabled:
        @cache()
        def cached_func(p):
            return bm.grad(p, **kwargs)
    else:  # pragma: no cover
        def cached_func(p):
            return bm.grad(p, **kwargs)

    def grad(p):
        return copy.deepcopy(cached_func(tuple(p)))

    return grad



import numpy as np
import scipy
import ionbench
import ionbench.utils.population_optimisers as pop_opt
import copy


# noinspection PyShadowingNames
def run(bm, x0=None, nGens=50, eta_cross=10, elitePercentage=0.066, popSize=50, debug=False):
    """
    Runs the genetic algorithm from Smirnov et al. 2020.

    Parameters
    ----------
    bm : Benchmarker
        A benchmarker to evaluate the performance of the optimisation algorithm.
    x0 : list, optional
        Initial parameter guess. Population is generated by randomly perturbing this initial guess +-50%, clamped to bounds if necessary. If x0=None (the default), then the population will be sampled using bm.sample().
    nGens : int, optional
        The number of generations to run the optimisation algorithm for. The default is 50.
    eta_cross : float, optional
        Crossover parameter. The default is 10.
    elitePercentage : float, optional
        The percentage of the population that are considered elites to move into the next generation. This will be multiplied by popSize and then rounded to the nearest integer. The default is 0.066.
    popSize : int, optional
        The size of the population in each generation. The default is 50.
    debug : bool, optional
        If True, debug information will be printed, reporting that status of the optimisation each generation. The default is False.

    Returns
    -------
    xbest : list
        The best parameters identified.

    """
    cost_func = ionbench.utils.cache.get_cached_cost(bm)

    eliteCount = int(np.round(popSize * elitePercentage))
    pop = pop_opt.get_pop(bm, x0, popSize, cost_func)
    gen = None
    for gen in range(nGens):
        elites = pop_opt.get_elites(pop, eliteCount)
        if debug:
            print("------------")
            print(f'Gen {gen}, Best cost: {elites[0].cost}')

        # Tournament selection
        pop = pop_opt.tournament_selection(pop)

        # Crossover SBX
        pop = pop_opt.sbx_crossover(pop, bm, cost_func, eta_cross)

        # Mutation
        for i in range(popSize):
            if np.random.rand() < 0.9:
                direc = np.random.rand(bm.n_parameters())
                direc = direc / np.linalg.norm(direc)
                mag = scipy.stats.cauchy.rvs(loc=0, scale=0.18)
                pop[i].x += mag * direc

        if debug:
            print(f'Finishing gen {gen}')

        # Find costs
        pop = pop_opt.find_pop_costs(pop)

        # Elitism
        pop = pop_opt.set_elites(pop, elites)

        if bm.is_converged():
            break

    if gen >= nGens - 1:
        bm.set_max_iter_flag()

    elites = pop_opt.get_elites(pop, 1)
    bm.evaluate()
    return elites[0].x


# noinspection PyUnusedLocal,PyShadowingNames
def get_modification(modNum=1):
    """
    modNum = 1 -> Smirnov2020

    Returns
    -------
    mod : modification
        Modification corresponding to inputted modNum. Default is modNum = 1, so Smirnov2020.

    """
    mod = ionbench.modification.Smirnov2020()
    return mod
